#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""A bot client might go here... eventually..."""
__all__ = ["BotBase", "StatelessBot"]

import abc
import asyncio
import datetime
import logging
import math
import time
import typing

from hikari import events
from hikari import gateway_entities
from hikari import guilds
from hikari import intents
from hikari.clients import configs
from hikari.clients import rest
from hikari.clients import runnable
from hikari.clients import shards
from hikari.internal import conversions
from hikari.internal import more_collections
from hikari.internal import more_typing
from hikari.state import dispatchers
from hikari.state import event_managers
from hikari.state import stateless

ShardClientT = typing.TypeVar("ShardClientT", bound=shards.ShardClient)
EventManagerT = typing.TypeVar("EventManagerT", bound=event_managers.EventManager)
RESTClientT = typing.TypeVar("RESTClientT", bound=rest.RESTClient)
BotConfigT = typing.TypeVar("BotConfigT", bound=configs.BotConfig)


class BotBase(
    typing.Generic[ShardClientT, RESTClientT, EventManagerT, BotConfigT],
    runnable.RunnableClient,
    dispatchers.EventDispatcher,
    abc.ABC,
):
    """An abstract base class for a bot implementation.

    This takes several generic parameter types in the following order:
    - ``ShardClientT`` - the implementation of
        :obj:`~hikari.clients.shard_clients.ShardClient` to use for shards.
    - ``RESTClientT`` - the implementation of
        :obj:`~hikari.clients.rest.RESTClient` to use for API calls.
    - ``EventManagerT`` - the implementation of
        :obj:`~hikari.state.event_managers.EventManager` to use for
        event management, translation, and dispatching.
    - ``BotConfigT`` - the implementation of
        :obj:`~hikari.clients.configs.BotConfig` to read component-specific
        details from.

    Parameters
    ----------
    config : :obj:`~hikari.clients.configs.BotConfig`
        The config object to use.
    """

    #: The config for this bot.
    #:
    #: :type: :obj:`~hikari.clients.configs.BotConfig`
    _config: BotConfigT

    #: The event manager for this bot.
    #:
    #: :type: an implementation instance of :obj:`~hikari.state.event_managers.EventManager`
    event_manager: EventManagerT

    #: The logger to use for this bot.
    #:
    #: :type: :obj:`~logging.Logger`
    logger: logging.Logger

    #: The REST HTTP client to use for this bot.
    #:
    #: :type: :obj:`~hikari.clients.rest.RESTClient`
    rest: RESTClientT

    #: Shards registered to this bot.
    #:
    #: These will be created once the bot has started execution.
    #:
    #: :type: :obj:`~typing.Mapping` [ :obj:`~int`, ? extends :obj:`~hikari.clients.shard_client.ShardClient` ]
    shards: typing.Mapping[int, ShardClientT]

    @abc.abstractmethod
    def __init__(self, config: configs.BotConfig) -> None:
        super().__init__(logging.getLogger(f"hikari.{type(self).__qualname__}"))
        self._config = config
        self.event_manager = self._create_event_manager()
        self.rest = self._create_rest(config)
        self.shards = more_collections.EMPTY_DICT

    @property
    def heartbeat_latency(self) -> float:
        """Average heartbeat latency for all valid shards.

        This will return a mean of all the heartbeat intervals for all shards
        with a valid heartbeat latency that are in the
        :obj:`~hikari.clients.shard_clients.ShardState.READY` state.

        If no shards are in this state, this will return ``float('nan')``
        instead.

        Returns
        -------
        :obj:`~float`
            The mean latency for all ``READY`` shards that have sent at least
            one acknowledged ``HEARTBEAT`` payload. If there is not at least
            one shard that meets this criteria, this will instead return
            ``float('nan')``.
        """
        latencies = []
        for shard in self.shards.values():
            if not math.isnan(shard.heartbeat_latency):
                latencies.append(shard.heartbeat_latency)

        return sum(latencies) / len(latencies) if latencies else float("nan")

    @property
    def total_disconnect_count(self) -> int:
        """Total number of times any shard has disconnected.

        Returns
        -------
        :obj:`int`
            Total disconnect count.
        """
        return sum(s.disconnect_count for s in self.shards.values())

    @property
    def total_reconnect_count(self) -> int:
        """Total number of times any shard has reconnected.

        Returns
        -------
        :obj:`int`
            Total reconnect count.
        """
        return sum(s.reconnect_count for s in self.shards.values())

    @property
    def intents(self) -> typing.Optional[intents.Intent]:
        """Intent values that any shard connections will be using.

        Returns
        -------
        :obj:`~hikari.intents.Intent`, optional
            A :obj:`~enum.IntFlag` enum containing each intent that is set. If
            intents are not being used at all, then this will return
            :obj:`~None` instead.
        """
        return self._config.intents

    @property
    def version(self) -> float:
        """Version being used for the gateway API.

        Returns
        -------
        :obj:`~int`
            The API version being used.
        """
        return self._config.gateway_version

    async def start(self):
        if self.shards:
            raise RuntimeError("Bot is already running.")

        gateway_bot = await self.rest.fetch_gateway_bot()

        self.logger.info(
            "you have sent an IDENTIFY %s time(s) before now, and have %s remaining. This will reset at %s.",
            gateway_bot.session_start_limit.total - gateway_bot.session_start_limit.remaining,
            gateway_bot.session_start_limit.remaining,
            datetime.datetime.now() + gateway_bot.session_start_limit.reset_after,
        )

        shard_count = self._config.shard_count if self._config.shard_count else gateway_bot.shard_count
        shard_ids = self._config.shard_ids if self._config.shard_ids else [*range(shard_count)]
        url = gateway_bot.url

        self.logger.info("will connect shards to %s", url)

        shards = {}
        for shard_id in shard_ids:
            shard = self._create_shard(shard_id, shard_count, url, self._config, self.event_manager)
            shards[shard_id] = shard

        self.shards = shards

        self.logger.info("starting %s", conversions.pluralize(len(self.shards), "shard"))

        start_time = time.perf_counter()

        for i, shard_id in enumerate(self.shards):
            if i > 0:
                self.logger.info("idling for 5 seconds to avoid an invalid session")
                await asyncio.sleep(5)

            shard_obj = self.shards[shard_id]
            await shard_obj.start()

        finish_time = time.perf_counter()

        self.logger.info("started %s shard(s) in approx %.2fs", len(self.shards), finish_time - start_time)

        if self.event_manager is not None:
            await self.dispatch_event(events.StartedEvent())

    async def close(self) -> None:
        try:
            if self.shards:
                self.logger.info("stopping %s shard(s)", len(self.shards))
                start_time = time.perf_counter()
                try:
                    await self.dispatch_event(events.StoppingEvent())
                    await asyncio.gather(*(shard_obj.close() for shard_obj in self.shards.values()))
                finally:
                    finish_time = time.perf_counter()
                    self.logger.info("stopped %s shard(s) in approx %.2fs", len(self.shards), finish_time - start_time)
                    await self.dispatch_event(events.StoppedEvent())
        finally:
            await self.rest.close()

    async def join(self) -> None:
        await asyncio.gather(*(shard_obj.join() for shard_obj in self.shards.values()))

    def add_listener(
        self, event_type: typing.Type[dispatchers.EventT], callback: dispatchers.EventCallbackT
    ) -> dispatchers.EventCallbackT:
        return self.event_manager.event_dispatcher.add_listener(event_type, callback)

    def remove_listener(
        self, event_type: typing.Type[dispatchers.EventT], callback: dispatchers.EventCallbackT
    ) -> dispatchers.EventCallbackT:
        return self.event_manager.event_dispatcher.remove_listener(event_type, callback)

    def wait_for(
        self,
        event_type: typing.Type[dispatchers.EventT],
        *,
        timeout: typing.Optional[float],
        predicate: dispatchers.PredicateT,
    ) -> more_typing.Future:
        return self.event_manager.event_dispatcher.wait_for(event_type, timeout=timeout, predicate=predicate)

    def dispatch_event(self, event: events.HikariEvent) -> more_typing.Future[typing.Any]:
        return self.event_manager.event_dispatcher.dispatch_event(event)

    async def update_presence(
        self,
        *,
        status: guilds.PresenceStatus = ...,
        activity: typing.Optional[gateway_entities.Activity] = ...,
        idle_since: typing.Optional[datetime.datetime] = ...,
        is_afk: bool = ...,
    ) -> None:
        """Update the presence of the user for all shards.

        This will only update arguments that you explicitly specify a value for.
        Any arguments that you do not explicitly provide some value for will
        not be changed.

        Warning
        -------
        This will only apply to connected shards.

        Notes
        -----
        If you wish to update a presence for a specific shard, you can do this
        by using the ``shards`` :obj:`~typing.Mapping` to find the shard you
        wish to update.

        Parameters
        ----------
        status : :obj:`~hikari.guilds.PresenceStatus`
            If specified, the new status to set.
        activity : :obj:`~hikari.gateway_entities.GatewayActivity`, optional
            If specified, the new activity to set.
        idle_since : :obj:`~datetime.datetime`, optional
            If specified, the time to show up as being idle since,
            or :obj:`~None` if not applicable.
        is_afk : :obj:`~bool`
            If specified, :obj:`~True` if the user should be marked as AFK,
            or :obj:`~False` otherwise.
        """
        await asyncio.gather(
            *(
                s.update_presence(status=status, activity=activity, idle_since=idle_since, is_afk=is_afk)
                for s in self.shards.values()
                if s.connection_state in (shards.ShardState.WAITING_FOR_READY, shards.ShardState.READY)
            )
        )

    @classmethod
    @abc.abstractmethod
    def _create_shard(
        cls, shard_id: int, shard_count: int, url: str, config: BotConfigT, event_manager: EventManagerT,
    ) -> ShardClientT:
        """Return a new shard for the given parameters.

        Parameters
        ----------
        shard_id : :obj:`~int`
            The shard ID to use.
        shard_count : :obj:`~int`
            The shard count to use.
        url : :obj:`~str`
            The gateway URL to connect to.
        config : :obj:`~hikari.clients.configs.BotConfig`
            The bot config to use.
        event_manager :obj:`~hikari.state.event_managers.EventManager`
            The event manager to use.

        Returns
        -------
        :obj:`~hikari.clients.shard_clients.ShardClient`
            The shard client implementation to use for the given shard ID.

        Notes
        -----
        The ``shard_id`` and ``shard_count`` may be set within the ``config``
        object passed, but any conforming implementations are expected to
        use the value passed in the ``shard_id` and ``shard_count`` parameters
        regardless. Failure to do so may result in an invalid sharding
        configuration being used.

        """

    @classmethod
    @abc.abstractmethod
    def _create_rest(cls, config: BotConfigT) -> RESTClientT:
        """Return a new REST client from the given configuration.

        Parameters
        ----------
        config : :obj:`~hikari.clients.configs.BotConfig`
            The bot config to use.

        Returns
        -------
        :obj:`~hikari.clients.rest.RESTClient`
            The REST client to use.

        """

    @classmethod
    @abc.abstractmethod
    def _create_event_manager(cls):
        """Return a new instance of an event manager implementation.

        Returns
        -------
        :obj:`~hikari.state.event_managers.EventManager`
            The event manager to use internally.
        """


class StatelessBot(
    BotBase[shards.ShardClientImpl, rest.RESTClient, stateless.StatelessEventManagerImpl, configs.BotConfig,]
):
    """Bot client without any state internals."""

    def __init__(self, config=configs.BotConfig) -> None:
        super().__init__(config)

    @classmethod
    def _create_shard(
        cls,
        shard_id: int,
        shard_count: int,
        url: str,
        config: configs.BotConfig,
        event_manager: stateless.StatelessEventManagerImpl,
    ) -> shards.ShardClientImpl:
        return shards.ShardClientImpl(
            shard_id=shard_id,
            shard_count=shard_count,
            config=config,
            raw_event_consumer_impl=event_manager,
            url=url,
            dispatcher=event_manager.event_dispatcher,
        )

    @classmethod
    def _create_rest(cls, config: BotConfigT) -> rest.RESTClient:
        return rest.RESTClient(config)

    @classmethod
    def _create_event_manager(cls) -> EventManagerT:
        return stateless.StatelessEventManagerImpl()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Special data structures and utilities.

|internal|
"""

__all__ = [
    "EMPTY_SEQUENCE",
    "EMPTY_SET",
    "EMPTY_COLLECTION",
    "EMPTY_DICT",
    "EMPTY_GENERATOR_EXPRESSION",
    "WeakKeyDictionary",
]

import types
import typing

import weakref

_T = typing.TypeVar("_T")
_K = typing.TypeVar("_K", bound=typing.Hashable)
_V = typing.TypeVar("_V")

EMPTY_SEQUENCE: typing.Final[typing.Sequence[_T]] = tuple()
EMPTY_SET: typing.Final[typing.AbstractSet[_T]] = frozenset()
EMPTY_COLLECTION: typing.Final[typing.Collection[_T]] = tuple()
EMPTY_DICT: typing.Final[typing.Mapping[_K, _V]] = types.MappingProxyType({})
EMPTY_GENERATOR_EXPRESSION: typing.Final[typing.Iterator[_T]] = (_ for _ in EMPTY_COLLECTION)


class WeakKeyDictionary(typing.Generic[_K, _V], weakref.WeakKeyDictionary, typing.MutableMapping[_K, _V]):
    """A dictionary that has weak references to the keys.

    This is a type-safe version of :obj:`~weakref.WeakKeyDictionary` which
    is subscriptable.

    Example
    -------
    .. code-block:: python

        @attr.s(auto_attribs=True)
        class Commands:
            instances: Set[Command]
            aliases: WeakKeyDictionary[Command, str]

    """

    __slots__ = ()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""An internal marshalling utility used by internal API components.

Warning
-------
You should not change anything in this file, if you do, you will likely get
unexpected behaviour elsewhere.

|internal|
"""
__all__ = [
    "RAISE",
    "dereference_handle",
    "attrib",
    "marshallable",
    "HIKARI_ENTITY_MARSHALLER",
    "HikariEntityMarshaller",
    "Deserializable",
    "Serializable",
]

import importlib
import typing
import weakref

import attr

from hikari.internal import assertions
from hikari.internal import more_typing

_RAW_NAME_ATTR: typing.Final[str] = __name__ + "_RAW_NAME"
_SERIALIZER_ATTR: typing.Final[str] = __name__ + "_SERIALIZER"
_DESERIALIZER_ATTR: typing.Final[str] = __name__ + "_DESERIALIZER"
_TRANSIENT_ATTR: typing.Final[str] = __name__ + "_TRANSIENT"
_IF_UNDEFINED: typing.Final[str] = __name__ + "IF_UNDEFINED"
_IF_NONE: typing.Final[str] = __name__ + "_IF_NONE"
_PASSED_THROUGH_SINGLETONS: typing.Final[typing.Sequence[bool]] = [False, True, None]
RAISE: typing.Final[typing.Any] = object()
EntityT = typing.TypeVar("EntityT", contravariant=True)


def dereference_handle(handle_string: str) -> typing.Any:
    """Parse a given handle string into an object reference.

    Parameters
    ----------
    handle_string : :obj:`~str`
        The handle to the object to refer to. This is in the format
        ``fully.qualified.module.name#object.attribute``. If no ``#`` is
        input, then the reference will be made to the module itself.

    Returns
    -------
    :obj:`~typing.Any`
        The thing that is referred to from this reference.

    Examples
    --------
    ``"collections#deque"``:
        Refers to :obj:`~collections.deque`
    ``"asyncio.tasks#Task"``:
        Refers to ``asyncio.tasks.Task``
    ``"hikari.net"``:
        Refers to :obj:`~hikari.net`
    ``"foo.bar#baz.bork.qux"``:
        Would refer to a theoretical ``qux`` attribute on a ``bork``
        attribute on a ``baz`` object in the ``foo.bar`` module.
    """
    if "#" not in handle_string:
        module, attribute_names = handle_string, ()
    else:
        module, _, attribute_string = handle_string.partition("#")
        attribute_names = attribute_string.split(".")

    obj = importlib.import_module(module)
    for attr_name in attribute_names:
        obj = getattr(obj, attr_name)

    return weakref.proxy(obj)


def attrib(
    *,
    # Mandatory! We do not want to rely on type annotations alone, as they will
    # break if we use __future__.annotations anywhere. If we relied on the
    # field type, that would work, but attrs doesn't let us supply field.type
    # as an attr.ib() kwargs AND use type hints at the same time, and without
    # type hints, the library loses the ability to be type checked properly
    # anymore, so we have to pass this explicitly regardless.
    deserializer: typing.Optional[typing.Callable[[typing.Any], typing.Any]] = None,
    if_none: typing.Union[typing.Callable[[], typing.Any], None, type(RAISE)] = RAISE,
    if_undefined: typing.Union[typing.Callable[[], typing.Any], None, type(RAISE)] = RAISE,
    raw_name: typing.Optional[str] = None,
    transient: bool = False,
    serializer: typing.Optional[typing.Callable[[typing.Any], typing.Any]] = None,
    **kwargs,
) -> attr.Attribute:
    """Create an :func:`attr.ib` with marshaller metadata attached.

    Parameters
    ----------
    deserializer : :obj:`~typing.Callable` [ [ :obj:`~typing.Any` ], :obj:`~typing.Any` ], optional
        The deserializer to use to deserialize raw elements.
    raw_name : :obj:`~str`, optional
        The raw name of the element in its raw serialized form. If not provided,
        then this will use the field's default name later.
    transient : :obj:`~bool`
        If :obj:`~True`, the field is marked as transient, meaning it will not be
        serialized. Defaults to :obj:`~False`.
    if_none
        Either a default factory function called to get the default for when
        this field is :obj:`~None` or one of :obj:`~None`, :obj:`~False` or
        :obj:`~True` to specify that this should default to the given singleton.
        Will raise an exception when :obj:`~None` is received for this field
        later if this isn't specified.
    if_undefined
        Either a default factory function called to get the default for when
        this field isn't defined or one of :obj:`~None`, :obj:`~False` or
        :obj:`~True` to specify that this should default to the given singleton.
        Will raise an exception when this field is undefined later on if this
        isn't specified.
    serializer : :obj:`~typing.Callable` [ [ :obj:`~typing.Any` ], :obj:`~typing.Any` ], optional
        The serializer to use. If not specified, then serializing the entire
        class that this attribute is in will trigger a :obj:`~TypeError`
        later.
    **kwargs :
        Any kwargs to pass to :func:`attr.ib`.

    Returns
    -------
    :obj:`~typing.Any`
        The result of :func:`attr.ib` internally being called with additional
        metadata.
    """
    # Sphinx decides to be really awkward and inject the wrong default values
    # by default. Not helpful when it documents non-optional shit as defaulting
    # to None. Hack to fix this seems to be to turn on autodoc's
    # typing.TYPE_CHECKING mode, and then if that is enabled, always return
    # some dummy class that has a repr that returns a literal "..." string.
    if typing.TYPE_CHECKING:
        return type("Literal", (), {"__repr__": lambda *_: "..."})()

    metadata = kwargs.pop("metadata", {})
    metadata[_RAW_NAME_ATTR] = raw_name
    metadata[_SERIALIZER_ATTR] = serializer
    metadata[_DESERIALIZER_ATTR] = deserializer
    metadata[_IF_NONE] = if_none
    metadata[_IF_UNDEFINED] = if_undefined
    metadata[_TRANSIENT_ATTR] = transient

    attribute = attr.ib(**kwargs, metadata=metadata)
    # Fool pylint into thinking this is any type.
    return typing.cast(typing.Any, attribute)


def _not_implemented(op, name):
    def error(*_, **__) -> typing.NoReturn:
        raise NotImplementedError(f"Field {name} does not support operation {op}")

    return error


def _default_validator(value: typing.Any):
    assertions.assert_that(
        value is RAISE or value in _PASSED_THROUGH_SINGLETONS or callable(value),
        message=(
            "Invalid default factory passed for `if_undefined` or `if_none`; "
            f"expected a callable or one of the 'passed through singletons' but got {value}."
        ),
        error_type=RuntimeError,
    )


class _AttributeDescriptor:
    __slots__ = (
        "raw_name",
        "field_name",
        "constructor_name",
        "if_none",
        "if_undefined",
        "is_transient",
        "deserializer",
        "serializer",
    )

    def __init__(
        self,
        raw_name: str,
        field_name: str,
        constructor_name: str,
        if_none: typing.Union[typing.Callable[..., typing.Any], None, type(RAISE)],
        if_undefined: typing.Union[typing.Callable[..., typing.Any], None, type(RAISE)],
        is_transient: bool,
        deserializer: typing.Callable[[typing.Any], typing.Any],
        serializer: typing.Callable[[typing.Any], typing.Any],
    ) -> None:
        _default_validator(if_undefined)
        _default_validator(if_none)
        self.raw_name = raw_name
        self.field_name = field_name
        self.constructor_name = constructor_name
        self.if_none = if_none
        self.if_undefined = if_undefined
        self.is_transient = is_transient  # Do not serialize
        self.deserializer = deserializer
        self.serializer = serializer


class _EntityDescriptor:
    __slots__ = ("entity_type", "attribs")

    def __init__(self, entity_type: typing.Type, attribs: typing.Collection[_AttributeDescriptor]) -> None:
        self.entity_type = entity_type
        self.attribs = tuple(attribs)


def _construct_attribute_descriptor(field: attr.Attribute) -> _AttributeDescriptor:
    raw_name = typing.cast(str, field.metadata.get(_RAW_NAME_ATTR) or field.name)
    field_name = typing.cast(str, field.name)

    constructor_name = field_name

    # Attrs strips leading underscores for generated __init__ methods.
    while constructor_name.startswith("_"):
        constructor_name = constructor_name[1:]

    return _AttributeDescriptor(
        raw_name=raw_name,
        field_name=field_name,
        constructor_name=constructor_name,
        if_none=field.metadata[_IF_NONE],
        if_undefined=field.metadata[_IF_UNDEFINED],
        is_transient=field.metadata[_TRANSIENT_ATTR],
        deserializer=field.metadata[_DESERIALIZER_ATTR] or _not_implemented("deserialize", field_name),
        serializer=field.metadata[_SERIALIZER_ATTR] or _not_implemented("serialize", field_name),
    )


def _construct_entity_descriptor(entity: typing.Any) -> _EntityDescriptor:
    assertions.assert_that(
        hasattr(entity, "__attrs_attrs__"),
        f"{entity.__module__}.{entity.__qualname__} is not an attr class",
        error_type=TypeError,
    )

    return _EntityDescriptor(entity, [_construct_attribute_descriptor(field) for field in attr.fields(entity)])


class HikariEntityMarshaller:
    """Hikari's utility to manage automated serialization and deserialization.

    It can deserialize and serialize any internal components that that are
    decorated with the :obj:`~marshallable` decorator, and that are
    :func:`attr.s` classes using fields with the :obj:`~attrib` function call
    descriptor.
    """

    __slots__ = ("_registered_entities",)

    def __init__(self) -> None:
        self._registered_entities: typing.MutableMapping[typing.Type, _EntityDescriptor] = {}

    def register(self, cls: typing.Type[EntityT]) -> typing.Type[EntityT]:
        """Register an attrs type for fast future deserialization.

        Parameters
        ----------
        cls : :obj:`~typing.Type` [ :obj:`~typing.Any` ]
            The type to register.

        Returns
        -------
        :obj:`~typing.Type` [ :obj:`~typing.Any` ]
            The input argument. This enables this to be used as a decorator if
            desired.

        Raises
        ------
        :obj:`~TypeError`
            If the class is not an :obj:`~attr.s` class.
        """
        entity_descriptor = _construct_entity_descriptor(cls)
        self._registered_entities[cls] = entity_descriptor
        return cls

    def deserialize(self, raw_data: more_typing.JSONObject, target_type: typing.Type[EntityT]) -> EntityT:
        """Deserialize a given raw data item into the target type.

        Parameters
        ----------
        raw_data : :obj:`~typing.Mapping` [ :obj:`~str`, :obj:`~typing.Any` ]
            The raw data to deserialize.
        target_type : :obj:`~typing.Type` [ :obj:`~typing.Any` ]
            The type to deserialize to.

        Returns
        -------
        :obj:`~typing.Any`
            The deserialized instance.

        Raises
        ------
        :obj:`~LookupError`
            If the entity is not registered.
        :obj:`~AttributeError`
            If the field is not optional, but the field was not present in the
            raw payload, or it was present, but it was assigned :obj:`~None`.
        :obj:`~TypeError`
            If the deserialization call failed for some reason.
        """
        try:
            descriptor = self._registered_entities[target_type]
        except KeyError:
            raise LookupError(f"No registered entity {target_type.__module__}.{target_type.__qualname__}")

        kwargs = {}

        for a in descriptor.attribs:
            kwarg_name = a.constructor_name

            if a.raw_name not in raw_data:
                if a.if_undefined is RAISE:
                    raise AttributeError(
                        "Failed to deserialize data to instance of "
                        f"{target_type.__module__}.{target_type.__qualname__} due to required field {a.field_name} "
                        f"(from raw key {a.raw_name}) not being included in the input payload\n\n{raw_data}"
                    )
                if a.if_undefined in _PASSED_THROUGH_SINGLETONS:
                    kwargs[kwarg_name] = a.if_undefined
                else:
                    kwargs[kwarg_name] = a.if_undefined()
                continue

            if (data := raw_data[a.raw_name]) is None:
                if a.if_none is RAISE:
                    raise AttributeError(
                        "Failed to deserialize data to instance of "
                        f"{target_type.__module__}.{target_type.__qualname__} due to non-nullable field {a.field_name}"
                        f" (from raw key {a.raw_name}) being `None` in the input payload\n\n{raw_data}"
                    )
                if a.if_none in _PASSED_THROUGH_SINGLETONS:
                    kwargs[kwarg_name] = a.if_none
                else:
                    kwargs[kwarg_name] = a.if_none()
                continue

            try:
                # Use the deserializer if it is there, otherwise use the constructor of the type of the field.
                kwargs[kwarg_name] = a.deserializer(data) if a.deserializer else data
            except Exception as exc:
                raise TypeError(
                    "Failed to deserialize data to instance of "
                    f"{target_type.__module__}.{target_type.__qualname__} because marshalling failed on "
                    f"attribute {a.field_name} (passed to constructor as {kwarg_name})"
                ) from exc

        return target_type(**kwargs)

    def serialize(self, obj: typing.Optional[typing.Any]) -> more_typing.NullableJSONObject:
        """Serialize a given entity into a raw data item.

        Parameters
        ----------
        obj : :obj:`~typing.Any`, optional
            The entity to serialize.

        Returns
        -------
        :obj:`~typing.Mapping` [ :obj:`~str`, :obj:`~typing.Any` ], optional
            The serialized raw data item.

        Raises
        ------
        :obj:`~LookupError`
            If the entity is not registered.
        """
        if obj is None:
            return None

        input_type = type(obj)

        try:
            descriptor = self._registered_entities[input_type]
        except KeyError:
            raise LookupError(f"No registered entity {input_type.__module__}.{input_type.__qualname__}")

        raw_data = {}

        for a in descriptor.attribs:
            if a.is_transient:
                continue
            value = getattr(obj, a.field_name)
            if value is not None:
                raw_data[a.raw_name] = a.serializer(value)

        return raw_data


HIKARI_ENTITY_MARSHALLER = HikariEntityMarshaller()


def marshallable(*, marshaller: HikariEntityMarshaller = HIKARI_ENTITY_MARSHALLER):
    """Create a decorator for a class to make it into an :obj:`~attr.s` class.

    Parameters
    ----------
    marshaller : :obj:`~HikariEntityMarshaller`
        If specified, this should be an instance of a marshaller to use. For
        most internal purposes, you want to not specify this, since it will
        then default to the hikari-global marshaller instead. This is
        useful, however, for testing and for external usage.

    Returns
    -------
    ``decorator(T) -> T``
        A decorator to decorate a class with.

    Notes
    -----
    The ``auto_attribs`` functionality provided by :obj:`~attr.s` is not
    supported by this marshaller utility. Do not attempt to use it!

    Example
    -------

    .. code-block:: python

        @attrs()
        class MyEntity:
            id: int = attrib(deserializer=int, serializer=str)
            password: str = attrib(deserializer=int, transient=True)
            ...

    """

    def decorator(cls):
        return marshaller.register(cls)

    return decorator


class Deserializable:
    """Mixin that enables the class to be deserialized from a raw entity."""

    __slots__ = ()

    @classmethod
    def deserialize(cls: typing.Type[more_typing.T_contra], payload: more_typing.JSONType) -> more_typing.T_contra:
        """Deserialize the given payload into the object.

        Parameters
        ----------
        payload
            The payload to deserialize into the object.
        """
        return HIKARI_ENTITY_MARSHALLER.deserialize(payload, cls)


class Serializable:
    """Mixin that enables an instance of the class to be serialized."""

    __slots__ = ()

    def serialize(self: more_typing.T_contra) -> more_typing.JSONType:
        """Serialize this instance into a naive value."""
        return HIKARI_ENTITY_MARSHALLER.serialize(self)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Assertions of things.

These are functions that validate a value, and are expected to return the value
on success but error on any failure. This allows for quick checking of
conditions that might break the function or cause it to misbehave.

|internal|
"""
__all__ = [
    "assert_that",
    "assert_not_none",
    "assert_in_range",
    "assert_is_int_power",
]

import math
import typing

ValueT = typing.TypeVar("ValueT")
BaseTypeInstanceT = typing.TypeVar("BaseTypeInstanceT")


def assert_that(condition: bool, message: str = None, error_type: type = ValueError) -> None:
    """If the given condition is falsified, raise a :obj:`~ValueError`.

    Will be raised with the optional description if provided.
    """
    if not condition:
        raise error_type(message or "condition must not be False")


def assert_not_none(value: ValueT, message: typing.Optional[str] = None) -> ValueT:
    """If the given value is :obj:`~None`, raise a :obj:`~ValueError`.

    Will be raised with the optional description if provided.
    """
    if value is None:
        raise ValueError(message or "value must not be None")
    return value


def assert_in_range(value, min_inclusive, max_inclusive, name: str = None):
    """If a value is not in the range [min, max], raise a :obj:`~ValueError`."""
    if not (min_inclusive <= value <= max_inclusive):
        name = name or "The value"
        raise ValueError(f"{name} must be in the inclusive range of {min_inclusive} and {max_inclusive}")


def assert_is_int_power(value: int, power: int) -> bool:
    """If a value is not a power the given int, raise :obj:`~ValueError`."""
    logarithm = math.log(value, power)
    assert_that(logarithm.is_integer(), f"value must be an integer power of {power}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""The logic for handling requests to channel endpoints."""

__all__ = ["RESTChannelComponent"]

import abc
import asyncio
import datetime
import typing

from hikari import bases
from hikari import channels as _channels
from hikari import embeds as _embeds
from hikari import guilds
from hikari import invites
from hikari import media
from hikari import messages as _messages
from hikari import permissions as _permissions
from hikari import users
from hikari import webhooks
from hikari.clients.rest import base
from hikari.internal import allowed_mentions
from hikari.internal import assertions
from hikari.internal import conversions
from hikari.internal import more_typing
from hikari.internal import pagination


class RESTChannelComponent(base.BaseRESTComponent, abc.ABC):  # pylint: disable=W0223
    """The REST client component for handling requests to channel endpoints."""

    async def fetch_channel(self, channel: bases.Hashable[_channels.Channel]) -> _channels.Channel:
        """Get an up to date channel object from a given channel object or ID.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object ID of the channel to look up.

        Returns
        -------
        :obj:`~hikari.channels.Channel`
            The channel object that has been found.

        Raises
        ------
        :obj:`~hikari.errors.BadRequestHTTPError`
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you don't have access to the channel.
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the channel does not exist.
        """
        payload = await self._session.get_channel(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel))
        )
        return _channels.deserialize_channel(payload)

    async def update_channel(
        self,
        channel: bases.Hashable[_channels.Channel],
        *,
        name: str = ...,
        position: int = ...,
        topic: str = ...,
        nsfw: bool = ...,
        bitrate: int = ...,
        user_limit: int = ...,
        rate_limit_per_user: typing.Union[int, datetime.timedelta] = ...,
        permission_overwrites: typing.Sequence[_channels.PermissionOverwrite] = ...,
        parent_category: typing.Optional[bases.Hashable[_channels.GuildCategory]] = ...,
        reason: str = ...,
    ) -> _channels.Channel:
        """Update one or more aspects of a given channel ID.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The channel ID to update.
        name : :obj:`~str`
            If specified, the new name for the channel. This must be
            inclusively between ``1`` and ``100`` characters in length.
        position : :obj:`~int`
            If specified, the position to change the channel to.
        topic : :obj:`~str`
            If specified, the topic to set. This is only applicable to
            text channels. This must be inclusively between ``0`` and ``1024``
            characters in length.
        nsfw : :obj:`~bool`
            Mark the channel as being not safe for work (NSFW) if :obj:`~True`.
            If :obj:`~False` or unspecified, then the channel is not marked as
            NSFW. Will have no visible effect for non-text guild channels.
        rate_limit_per_user : :obj:`~typing.Union` [ :obj:`~int`, :obj:`~datetime.timedelta` ]
            If specified, the time delta of seconds  the user has to wait
            before sending another message. This will not apply to bots, or to
            members with ``MANAGE_MESSAGES`` or ``MANAGE_CHANNEL`` permissions.
            This must be inclusively between ``0`` and ``21600`` seconds.
        bitrate : :obj:`~int`
            If specified, the bitrate in bits per second allowable for the
            channel. This only applies to voice channels and must be inclusively
            between ``8000`` and ``96000`` for normal servers or ``8000`` and
            ``128000`` for VIP servers.
        user_limit : :obj:`~int`
            If specified, the new max number of users to allow in a voice
            channel. This must be between ``0`` and ``99`` inclusive, where
            ``0`` implies no limit.
        permission_overwrites : :obj:`~typing.Sequence` [ :obj:`~hikari.channels.PermissionOverwrite` ]
            If specified, the new list of permission overwrites that are
            category specific to replace the existing overwrites with.
        parent_category : :obj:`~typing.Union` [ :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ], optional
            If specified, the new parent category ID to set for the channel,
            pass :obj:`~None` to unset.
        reason : :obj:`~str`
            If specified, the audit log reason explaining why the operation
            was performed.

        Returns
        -------
        :obj:`~hikari.channels.Channel`
            The channel object that has been modified.

        Raises
        ------
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the channel does not exist.
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you lack the permission to make the change.
        :obj:`~hikari.errors.BadRequestHTTPError`
            If you provide incorrect options for the corresponding channel type
            (e.g. a ``bitrate`` for a text channel).
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        """
        payload = await self._session.modify_channel(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
            name=name,
            position=position,
            topic=topic,
            nsfw=nsfw,
            bitrate=bitrate,
            user_limit=user_limit,
            rate_limit_per_user=(
                int(rate_limit_per_user.total_seconds())
                if isinstance(rate_limit_per_user, datetime.timedelta)
                else rate_limit_per_user
            ),
            permission_overwrites=(
                [po.serialize() for po in permission_overwrites] if permission_overwrites is not ... else ...
            ),
            parent_id=(
                str(parent_category.id if isinstance(parent_category, bases.UniqueEntity) else int(parent_category))
                if parent_category is not ... and parent_category is not None
                else parent_category
            ),
            reason=reason,
        )
        return _channels.deserialize_channel(payload)

    async def delete_channel(self, channel: bases.Hashable[_channels.Channel]) -> None:
        """Delete the given channel ID, or if it is a DM, close it.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake` :obj:`~str` ]
            The object or ID of the channel to delete.

        Returns
        -------
        :obj:`~None`
            Nothing, unlike what the API specifies. This is done to maintain
            consistency with other calls of a similar nature in this API
            wrapper.

        Raises
        ------
        :obj:`~hikari.errors.BadRequestHTTPError`
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the channel does not exist.
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you do not have permission to delete the channel.

        Note
        ----
        Closing a DM channel won't raise an exception but will have no effect
        and "closed" DM channels will not have to be reopened to send messages
        in theme.

        Warning
        -------
        Deleted channels cannot be un-deleted.
        """
        await self._session.delete_close_channel(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel))
        )

    def fetch_messages_after(
        self,
        channel: bases.Hashable[_channels.Channel],
        *,
        after: typing.Union[datetime.datetime, bases.Hashable[_messages.Message]] = 0,
        limit: typing.Optional[int] = None,
    ) -> typing.AsyncIterator[_messages.Message]:
        """Return an async iterator that retrieves a channel's message history.

        This will return the message created after a given message object/ID or
        from the first message in the channel.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The ID of the channel to retrieve the messages from.
        limit : :obj:`~int`
            If specified, the maximum number of how many messages this iterator
            should return.
        after : :obj:`~typing.Union` [ :obj:`~datetime.datetime`, :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            A object or ID message. Only return messages sent AFTER this
            message if it's specified else this will return every message after
            (and including) the first message in the channel.

        Example
        -------
        .. code-block:: python

            async for message in client.fetch_messages_after(channel, after=9876543, limit=3232):
                if message.author.id in BLACKLISTED_USERS:
                    await client.ban_member(channel.guild_id,  message.author)

        Returns
        -------
        :obj:`~typing.AsyncIterator` [ :obj:`~hikari.messages.Message` ]
            An async iterator that retrieves the channel's message objects.

        Raises
        ------
        :obj:`~hikari.errors.BadRequestHTTPError`
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you lack permission to read the channel.
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the channel is not found, or the message
            provided for one of the filter arguments is not found.

        Note
        ----
        If you are missing the ``VIEW_CHANNEL`` permission, you will receive a
        :obj:`~hikari.errors.ForbiddenHTTPError`. If you are instead missing
        the ``READ_MESSAGE_HISTORY`` permission, you will always receive
        zero results, and thus an empty list will be returned instead.
        """
        if isinstance(after, datetime.datetime):
            after = str(bases.Snowflake.from_datetime(after))
        else:
            after = str(after.id if isinstance(after, bases.UniqueEntity) else int(after))
        return pagination.pagination_handler(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
            deserializer=_messages.Message.deserialize,
            direction="after",
            start=after,
            request=self._session.get_channel_messages,
            reversing=True,  # This is the only known endpoint where reversing is needed.
            limit=limit,
        )

    def fetch_messages_before(
        self,
        channel: bases.Hashable[_channels.Channel],
        *,
        before: typing.Union[datetime.datetime, bases.Hashable[_messages.Message], None] = None,
        limit: typing.Optional[int] = None,
    ) -> typing.AsyncIterator[_messages.Message]:
        """Return an async iterator that retrieves a channel's message history.

        This returns the message created after a given message object/ID or
        from the first message in the channel.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The ID of the channel to retrieve the messages from.
        limit : :obj:`~int`
            If specified, the maximum number of how many messages this iterator
            should return.
        before : :obj:`~typing.Union` [ :obj:`~datetime.datetime`, :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            A message object or ID. Only return messages sent BEFORE
            this message if this is specified else this will return every
            message before (and including) the most recent message in the
            channel.

        Example
        -------
        .. code-block:: python

            async for message in client.fetch_messages_before(channel, before=9876543, limit=1231):
                if message.content.lower().contains("delete this"):
                    await client.delete_message(channel, message)

        Returns
        -------
        :obj:`~typing.AsyncIterator` [ :obj:`~hikari.messages.Message` ]
            An async iterator that retrieves the channel's message objects.

        Raises
        ------
        :obj:`~hikari.errors.BadRequestHTTPError`
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you lack permission to read the channel.
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the channel is not found, or the message
            provided for one of the filter arguments is not found.

        Note
        ----
        If you are missing the ``VIEW_CHANNEL`` permission, you will receive a
        :obj:`~hikari.errors.ForbiddenHTTPError`. If you are instead missing
        the ``READ_MESSAGE_HISTORY`` permission, you will always receive
        zero results, and thus an empty list will be returned instead.
        """
        if isinstance(before, datetime.datetime):
            before = str(bases.Snowflake.from_datetime(before))
        elif before is not None:
            before = str(before.id if isinstance(before, bases.UniqueEntity) else int(before))
        return pagination.pagination_handler(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
            deserializer=_messages.Message.deserialize,
            direction="before",
            start=before,
            request=self._session.get_channel_messages,
            reversing=False,
            limit=limit,
        )

    async def fetch_messages_around(
        self,
        channel: bases.Hashable[_channels.Channel],
        around: typing.Union[datetime.datetime, bases.Hashable[_messages.Message]],
        *,
        limit: int = ...,
    ) -> typing.AsyncIterator[_messages.Message]:
        """Return an async iterator that retrieves up to 100 messages.

        This will return messages in order from newest to oldest, is based
        around the creation time of the supplied message object/ID and will
        include the given message if it still exists.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The ID of the channel to retrieve the messages from.
        around : :obj:`~typing.Union` [ :obj:`~datetime.datetime`, :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the message to get messages that were sent
            AROUND it in the provided channel, unlike ``before`` and ``after``,
            this argument is required and the provided message will also be
            returned if it still exists.
        limit : :obj:`~int`
            If specified, the maximum number of how many messages this iterator
            should return, cannot be more than `100`

        Example
        -------
        .. code-block:: python

            async for message in client.fetch_messages_around(channel, around=9876543, limit=42):
                if message.embeds and not message.author.is_bot:
                    await client.delete_message(channel, message)

        Returns
        -------
        :obj:`~typing.AsyncIterator` [ :obj:`~hikari.messages.Message` ]
            An async iterator that retrieves the found message objects.

        Raises
        ------
        :obj:`~hikari.errors.BadRequestHTTPError`
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you lack permission to read the channel.
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the channel is not found, or the message
            provided for one of the filter arguments is not found.

        Note
        ----
        If you are missing the ``VIEW_CHANNEL`` permission, you will receive a
        :obj:`~hikari.errors.ForbiddenHTTPError`. If you are instead missing
        the ``READ_MESSAGE_HISTORY`` permission, you will always receive
        zero results, and thus an empty list will be returned instead.
        """
        if isinstance(around, datetime.datetime):
            around = str(bases.Snowflake.from_datetime(around))
        else:
            around = str(around.id if isinstance(around, bases.UniqueEntity) else int(around))
        for payload in await self._session.get_channel_messages(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
            limit=limit,
            around=around,
        ):
            yield _messages.Message.deserialize(payload)

    async def fetch_message(
        self, channel: bases.Hashable[_channels.Channel], message: bases.Hashable[_messages.Message],
    ) -> _messages.Message:
        """Get a message from known channel that we can access.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the channel to get the message from.
        message : :obj:`~typing.Union` [ :obj:`~hikari.messages.Message`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the message to retrieve.

        Returns
        -------
        :obj:`~hikari.messages.Message`
            The found message object.

        Note
        ----
        This requires the ``READ_MESSAGE_HISTORY`` permission.

        Raises
        ------
        :obj:`~hikari.errors.BadRequestHTTPError`
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you lack permission to see the message.
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the channel or message is not found.
        """
        payload = await self._session.get_channel_message(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
            message_id=str(message.id if isinstance(message, bases.UniqueEntity) else int(message)),
        )
        return _messages.Message.deserialize(payload)

    async def create_message(
        self,
        channel: bases.Hashable[_channels.Channel],
        *,
        content: str = ...,
        nonce: str = ...,
        tts: bool = ...,
        files: typing.Collection[media.IO] = ...,
        embed: _embeds.Embed = ...,
        mentions_everyone: bool = True,
        user_mentions: typing.Union[typing.Collection[bases.Hashable[users.User]], bool] = True,
        role_mentions: typing.Union[typing.Collection[bases.Hashable[guilds.GuildRole]], bool] = True,
    ) -> _messages.Message:
        """Create a message in the given channel.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The channel or ID of the channel to send to.
        content : :obj:`~str`
            If specified, the message content to send with the message.
        nonce : :obj:`~str`
            If specified, an optional ID to send for opportunistic message
            creation. This doesn't serve any real purpose for general use,
            and can usually be ignored.
        tts : :obj:`~bool`
            If specified, whether the message will be sent as a TTS message.
        files : :obj:`~typing.Collection` [ ``hikari.media.IO`` ]
            If specified, this should be a list of inclusively between ``1`` and
            ``5`` IO like media objects, as defined in :mod:`hikari.media`.
        embed : :obj:`~hikari.embeds.Embed`
            If specified, the embed object to send with the message.
        mentions_everyone : :obj:`~bool`
            Whether ``@everyone`` and ``@here`` mentions should be resolved by
            discord and lead to actual pings, defaults to :obj:`~True`.
        user_mentions : :obj:`~typing.Union` [ :obj:`~typing.Collection` [ :obj:`~typing.Union` [ :obj:`~hikari.users.User`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ], :obj:`~bool` ]
            Either an array of user objects/IDs to allow mentions for,
            :obj:`~True` to allow all user mentions or :obj:`~False` to block all
            user mentions from resolving, defaults to :obj:`~True`.
        role_mentions : :obj:`~typing.Union` [ :obj:`~typing.Collection` [ :obj:`~typing.Union` [ :obj:`~hikari.guilds.GuildRole`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ] ], :obj:`~bool` ]
            Either an array of guild role objects/IDs to allow mentions for,
            :obj:`~True` to allow all role mentions or :obj:`~False` to block all
            role mentions from resolving, defaults to :obj:`~True`.

        Returns
        -------
        :obj:`~hikari.messages.Message`
            The created message object.

        Raises
        ------
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the channel is not found.
        :obj:`~hikari.errors.BadRequestHTTPError`
            This can be raised if the file is too large; if the embed exceeds
            the defined limits; if the message content is specified only and
            empty or greater than ``2000`` characters; if neither content, files
            or embed are specified.
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you lack permissions to send to this channel.
        :obj:`~ValueError`
            If more than 100 unique objects/entities are passed for
            ``role_mentions`` or ``user_mentions``.
        """
        payload = await self._session.create_message(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
            content=content,
            nonce=nonce,
            tts=tts,
            files=await asyncio.gather(*(media.safe_read_file(file) for file in files)) if files is not ... else ...,
            embed=embed.serialize() if embed is not ... else ...,
            allowed_mentions=allowed_mentions.generate_allowed_mentions(
                mentions_everyone=mentions_everyone, user_mentions=user_mentions, role_mentions=role_mentions
            ),
        )
        return _messages.Message.deserialize(payload)

    def safe_create_message(
        self,
        channel: bases.Hashable[_channels.Channel],
        *,
        content: str = ...,
        nonce: str = ...,
        tts: bool = ...,
        files: typing.Collection[media.IO] = ...,
        embed: _embeds.Embed = ...,
        mentions_everyone: bool = False,
        user_mentions: typing.Union[typing.Collection[bases.Hashable[users.User]], bool] = False,
        role_mentions: typing.Union[typing.Collection[bases.Hashable[guilds.GuildRole]], bool] = False,
    ) -> more_typing.Coroutine[_messages.Message]:
        """Create a message in the given channel with mention safety.

        This endpoint has the same signature as :attr:`create_message` with
        the only difference being that ``mentions_everyone``,
        ``user_mentions`` and ``role_mentions`` default to :obj:`~False`.
        """
        return self.create_message(
            channel=channel,
            content=content,
            nonce=nonce,
            tts=tts,
            files=files,
            embed=embed,
            mentions_everyone=mentions_everyone,
            user_mentions=user_mentions,
            role_mentions=role_mentions,
        )

    async def update_message(
        self,
        message: bases.Hashable[_messages.Message],
        channel: bases.Hashable[_channels.Channel],
        *,
        content: typing.Optional[str] = ...,
        embed: typing.Optional[_embeds.Embed] = ...,
        flags: int = ...,
        mentions_everyone: bool = True,
        user_mentions: typing.Union[typing.Collection[bases.Hashable[users.User]], bool] = True,
        role_mentions: typing.Union[typing.Collection[bases.Hashable[guilds.GuildRole]], bool] = True,
    ) -> _messages.Message:
        """Update the given message.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the channel to get the message from.
        message : :obj:`~typing.Union` [ :obj:`~hikari.messages.Message`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the message to edit.
        content : :obj:`~str`, optional
            If specified, the string content to replace with in the message.
            If :obj:`~None`, the content will be removed from the message.
        embed : :obj:`~hikari.embeds.Embed`, optional
            If specified, the embed to replace with in the message.
            If :obj:`~None`, the embed will be removed from the message.
        flags : :obj:`~hikari.messages.MessageFlag`
            If specified, the new flags for this message, while a raw int may
            be passed for this, this can lead to unexpected behaviour if it's
            outside the range of the MessageFlag int flag.
        mentions_everyone : :obj:`~bool`
            Whether ``@everyone`` and ``@here`` mentions should be resolved by
            discord and lead to actual pings, defaults to :obj:`~True`.
        user_mentions : :obj:`~typing.Union` [ :obj:`~typing.Collection` [ :obj:`~typing.Union` [ :obj:`~hikari.users.User`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ], :obj:`~bool` ]
            Either an array of user objects/IDs to allow mentions for,
            :obj:`~True` to allow all user mentions or :obj:`~False` to block all
            user mentions from resolving, defaults to :obj:`~True`.
        role_mentions : :obj:`~typing.Union` [ :obj:`~typing.Collection` [ :obj:`~typing.Union` [ :obj:`~hikari.guilds.GuildRole`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ] ], :obj:`~bool` ]
            Either an array of guild role objects/IDs to allow mentions for,
            :obj:`~True` to allow all role mentions or :obj:`~False` to block all
            role mentions from resolving, defaults to :obj:`~True`.

        Returns
        -------
        :obj:`~hikari.messages.Message`
            The edited message object.

        Raises
        ------
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the channel or message is not found.
        :obj:`~hikari.errors.BadRequestHTTPError`
            This can be raised if the embed exceeds the defined limits;
            if the message content is specified only and empty or greater
            than ``2000`` characters; if neither content, file or embed
            are specified.
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you try to edit ``content`` or ``embed`` or ``allowed_mentions`
            on a message you did not author.
            If you try to edit the flags on a message you did not author without
            the ``MANAGE_MESSAGES`` permission.
        :obj:`~ValueError`
            If more than 100 unique objects/entities are passed for
            ``role_mentions`` or ``user_mentions``.
        """
        payload = await self._session.edit_message(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
            message_id=str(message.id if isinstance(message, bases.UniqueEntity) else int(message)),
            content=content,
            embed=embed.serialize() if embed is not ... and embed is not None else embed,
            flags=flags,
            allowed_mentions=allowed_mentions.generate_allowed_mentions(
                mentions_everyone=mentions_everyone, user_mentions=user_mentions, role_mentions=role_mentions,
            ),
        )
        return _messages.Message.deserialize(payload)

    def safe_update_message(
        self,
        message: bases.Hashable[_messages.Message],
        channel: bases.Hashable[_channels.Channel],
        *,
        content: typing.Optional[str] = ...,
        embed: typing.Optional[_embeds.Embed] = ...,
        flags: int = ...,
        mentions_everyone: bool = False,
        user_mentions: typing.Union[typing.Collection[bases.Hashable[users.User]], bool] = False,
        role_mentions: typing.Union[typing.Collection[bases.Hashable[guilds.GuildRole]], bool] = False,
    ) -> typing.Coroutine[typing.Any, typing.Any, _messages.Message]:
        """Update a message in the given channel with mention safety.

        This endpoint has the same signature as :attr:`update_message` with
        the only difference being that ``mentions_everyone``,
        ``user_mentions`` and ``role_mentions`` default to :obj:`~False`.
        """
        return self.update_message(
            message=message,
            channel=channel,
            content=content,
            embed=embed,
            flags=flags,
            mentions_everyone=mentions_everyone,
            user_mentions=user_mentions,
            role_mentions=role_mentions,
        )

    async def delete_messages(
        self,
        channel: bases.Hashable[_channels.Channel],
        message: bases.Hashable[_messages.Message],
        *additional_messages: bases.Hashable[_messages.Message],
    ) -> None:
        """Delete a message in a given channel.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the channel to get the message from.
        message : :obj:`~typing.Union` [ :obj:`~hikari.messages.Message`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the message to delete.
        *additional_messages : :obj:`~typing.Union` [ :obj:`~hikari.messages.Message`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            Objects and/or IDs of additional messages to delete in the same
            channel, in total you can delete up to 100 messages in a request.

        Raises
        ------
        :obj:`~hikari.errors.BadRequestHTTPError`
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you did not author the message and are in a DM, or if you did
            not author the message and lack the ``MANAGE_MESSAGES``
            permission in a guild channel.
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the channel or message is not found.
        :obj:`~ValueError`
            If you try to delete over ``100`` messages in a single request.

        Note
        ----
        This can only be used on guild text channels.
        Any message IDs that do not exist or are invalid still add towards the
        total ``100`` max messages to remove. This can only delete messages
        that are newer than ``2`` weeks in age. If any of the messages ar
        older than ``2`` weeks then this call will fail.
        """
        if additional_messages:
            messages = list(
                # dict.fromkeys is used to remove duplicate entries that would cause discord to return an error.
                dict.fromkeys(
                    str(m.id if isinstance(m, bases.UniqueEntity) else int(m)) for m in (message, *additional_messages)
                )
            )
            assertions.assert_that(
                len(messages) <= 100, "Only up to 100 messages can be bulk deleted in a single request."
            )

            if len(messages) > 1:
                await self._session.bulk_delete_messages(
                    channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
                    messages=messages,
                )
                return None

        await self._session.delete_message(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
            message_id=str(message.id if isinstance(message, bases.UniqueEntity) else int(message)),
        )

    async def update_channel_overwrite(
        self,
        channel: bases.Hashable[_messages.Message],
        overwrite: typing.Union[_channels.PermissionOverwrite, users.User, guilds.GuildRole, bases.Snowflake, int],
        target_type: typing.Union[_channels.PermissionOverwriteType, str],
        *,
        allow: typing.Union[_permissions.Permission, int] = ...,
        deny: typing.Union[_permissions.Permission, int] = ...,
        reason: str = ...,
    ) -> None:
        """Edit permissions for a given channel.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.messages.Message`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the channel to edit permissions for.
        overwrite : :obj:`~typing.Union` [ :obj:`~hikari.channels.PermissionOverwrite`, :obj:`~hikari.guilds.GuildRole`, :obj:`~hikari.users.User`, :obj:`~hikari.entities.Snowflake` , :obj:`~int` ]
            The object or ID of the target member or role to  edit/create the
            overwrite for.
        target_type : :obj:`~typing.Union` [ :obj:`~hikari.channels.PermissionOverwriteType`, :obj:`~int` ]
            The type of overwrite, passing a raw string that's outside of the
            enum's range for this may lead to unexpected behaviour.
        allow : :obj:`~typing.Union` [ :obj:`~hikari.permissions.Permission`, :obj:`~int` ]
            If specified, the value of all permissions to set to be allowed,
            passing a raw integer for this may lead to unexpected behaviour.
        deny : :obj:`~typing.Union` [ :obj:`~hikari.permissions.Permission`, :obj:`~int` ]
            If specified, the value of all permissions to set to be denied,
            passing a raw integer for this may lead to unexpected behaviour.
        reason : :obj:`~str`
            If specified, the audit log reason explaining why the operation
            was performed.

        Raises
        ------
        :obj:`~hikari.errors.BadRequestHTTPError`
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the target channel or overwrite doesn't exist.
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you lack permission to do this.
        """
        await self._session.edit_channel_permissions(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
            overwrite_id=str(overwrite.id if isinstance(overwrite, bases.UniqueEntity) else int(overwrite)),
            type_=target_type,
            allow=allow,
            deny=deny,
            reason=reason,
        )

    async def fetch_invites_for_channel(
        self, channel: bases.Hashable[_channels.Channel]
    ) -> typing.Sequence[invites.InviteWithMetadata]:
        """Get invites for a given channel.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the channel to get invites for.

        Returns
        -------
        :obj:`~typing.Sequence` [ :obj:`~hikari.invites.InviteWithMetadata` ]
            A list of invite objects.

        Raises
        ------
        :obj:`~hikari.errors.BadRequestHTTPError`
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you lack the ``MANAGE_CHANNELS`` permission.
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the channel does not exist.
        """
        payload = await self._session.get_channel_invites(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel))
        )
        return [invites.InviteWithMetadata.deserialize(invite) for invite in payload]

    async def create_invite_for_channel(
        self,
        channel: bases.Hashable[_channels.Channel],
        *,
        max_age: typing.Union[int, datetime.timedelta] = ...,
        max_uses: int = ...,
        temporary: bool = ...,
        unique: bool = ...,
        target_user: bases.Hashable[users.User] = ...,
        target_user_type: typing.Union[invites.TargetUserType, int] = ...,
        reason: str = ...,
    ) -> invites.InviteWithMetadata:
        """Create a new invite for the given channel.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~datetime.timedelta`, :obj:`~str` ]
            The object or ID of the channel to create the invite for.
        max_age : :obj:`~int`
            If specified, the seconds time delta for the max age of the invite,
            defaults to ``86400`` seconds (``24`` hours).
            Set to ``0`` seconds to never expire.
        max_uses : :obj:`~int`
            If specified, the max number of uses this invite can have, or ``0``
            for unlimited (as per the default).
        temporary : :obj:`~bool`
            If specified, whether to grant temporary membership, meaning the
            user is kicked when their session ends unless they are given a role.
        unique : :obj:`~bool`
            If specified, whether to try to reuse a similar invite.
        target_user : :obj:`~typing.Union` [ :obj:`~hikari.users.User`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            If specified, the object or ID of the user this invite should
            target.
        target_user_type : :obj:`~typing.Union` [ :obj:`~hikari.invites.TargetUserType`, :obj:`~int` ]
            If specified, the type of target for this invite, passing a raw
            integer for this may lead to unexpected results.
        reason : :obj:`~str`
            If specified, the audit log reason explaining why the operation
            was performed.

        Returns
        -------
        :obj:`~hikari.invites.InviteWithMetadata`
            The created invite object.

        Raises
        ------
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you lack the ``CREATE_INSTANT_MESSAGES`` permission.
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the channel does not exist.
        :obj:`~hikari.errors.BadRequestHTTPError`
            If the arguments provided are not valid (e.g. negative age, etc).
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        """
        payload = await self._session.create_channel_invite(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
            max_age=int(max_age.total_seconds()) if isinstance(max_age, datetime.timedelta) else max_age,
            max_uses=max_uses,
            temporary=temporary,
            unique=unique,
            target_user=(
                str(target_user.id if isinstance(target_user, bases.UniqueEntity) else int(target_user))
                if target_user is not ...
                else ...
            ),
            target_user_type=target_user_type,
            reason=reason,
        )
        return invites.InviteWithMetadata.deserialize(payload)

    async def delete_channel_overwrite(
        self,
        channel: bases.Hashable[_channels.Channel],
        overwrite: typing.Union[_channels.PermissionOverwrite, guilds.GuildRole, users.User, bases.Snowflake, int],
    ) -> None:
        """Delete a channel permission overwrite for a user or a role.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the channel to delete the overwrite from.
        overwrite : :obj:`~typing.Union` [ :obj:`~hikari.channels.PermissionOverwrite`, :obj:`~hikari.guilds.GuildRole`, :obj:`~hikari.users.User`, :obj:`~hikari.entities.Snowflake`, :obj:int ]
            The ID of the entity this overwrite targets.

        Raises
        ------
        :obj:`~hikari.errors.BadRequestHTTPError`
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the overwrite or channel do not exist.
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you lack the ``MANAGE_ROLES`` permission for that channel.
        """
        await self._session.delete_channel_permission(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
            overwrite_id=str(overwrite.id if isinstance(overwrite, bases.UniqueEntity) else int(overwrite)),
        )

    async def trigger_typing(self, channel: bases.Hashable[_channels.Channel]) -> None:
        """Trigger the typing indicator for ``10`` seconds in a channel.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the channel to appear to be typing in.

        Raises
        ------
        :obj:`~hikari.errors.BadRequestHTTPError`
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the channel is not found.
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you are not able to type in the channel.
        """
        await self._session.trigger_typing_indicator(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel))
        )

    async def fetch_pins(
        self, channel: bases.Hashable[_channels.Channel]
    ) -> typing.Mapping[bases.Snowflake, _messages.Message]:
        """Get pinned messages for a given channel.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the channel to get messages from.

        Returns
        -------
        :obj:`~typing.Mapping` [ :obj:`~hikari.entities.Snowflake`, :obj:`~hikari.messages.Message` ]
            A list of message objects.

        Raises
        ------
        :obj:`~hikari.errors.BadRequestHTTPError`
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the channel is not found.
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you are not able to see the channel.

        Note
        ----
        If you are not able to see the pinned message (eg. you are missing
        ``READ_MESSAGE_HISTORY`` and the pinned message is an old message), it
        will not be returned.
        """
        payload = await self._session.get_pinned_messages(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel))
        )
        return {message.id: message for message in map(_messages.Message.deserialize, payload)}

    async def pin_message(
        self, channel: bases.Hashable[_channels.Channel], message: bases.Hashable[_messages.Message],
    ) -> None:
        """Add a pinned message to the channel.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the channel to pin a message to.
        message : :obj:`~typing.Union` [ :obj:`~hikari.messages.Message`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the message to pin.

        Raises
        ------
        :obj:`~hikari.errors.BadRequestHTTPError`
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you lack the ``MANAGE_MESSAGES`` permission.
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the message or channel do not exist.
        """
        await self._session.add_pinned_channel_message(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
            message_id=str(message.id if isinstance(message, bases.UniqueEntity) else int(message)),
        )

    async def unpin_message(
        self, channel: bases.Hashable[_channels.Channel], message: bases.Hashable[_messages.Message],
    ) -> None:
        """Remove a pinned message from the channel.

        This will only unpin the message, not delete it.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The ID of the channel to remove a pin from.
        message : :obj:`~typing.Union` [ :obj:`~hikari.messages.Message`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the message to unpin.

        Raises
        ------
        :obj:`~hikari.errors.BadRequestHTTPError`
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you lack the ``MANAGE_MESSAGES`` permission.
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the message or channel do not exist.
        """
        await self._session.delete_pinned_channel_message(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
            message_id=str(message.id if isinstance(message, bases.UniqueEntity) else int(message)),
        )

    async def create_webhook(
        self,
        channel: bases.Hashable[_channels.GuildChannel],
        name: str,
        *,
        avatar_data: conversions.FileLikeT = ...,
        reason: str = ...,
    ) -> webhooks.Webhook:
        """Create a webhook for a given channel.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.channels.GuildChannel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the channel for webhook to be created in.
        name : :obj:`~str`
            The webhook's name string.
        avatar_data : ``hikari.internal.conversions.FileLikeT``
            If specified, the avatar image data.
        reason : :obj:`~str`
            If specified, the audit log reason explaining why the operation
            was performed.

        Returns
        -------
        :obj:`~hikari.webhooks.Webhook`
            The newly created webhook object.

        Raises
        ------
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the channel is not found.
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you either lack the ``MANAGE_WEBHOOKS`` permission or
            can not see the given channel.
        :obj:`~hikari.errors.BadRequestHTTPError`
            If the avatar image is too big or the format is invalid.
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        """
        payload = await self._session.create_webhook(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
            name=name,
            avatar=conversions.get_bytes_from_resource(avatar_data) if avatar_data is not ... else ...,
            reason=reason,
        )
        return webhooks.Webhook.deserialize(payload)

    async def fetch_channel_webhooks(
        self, channel: bases.Hashable[_channels.GuildChannel]
    ) -> typing.Sequence[webhooks.Webhook]:
        """Get all webhooks from a given channel.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.channels.GuildChannel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the guild channel to get the webhooks from.

        Returns
        -------
        :obj:`~typing.Sequence` [ :obj:`~hikari.webhooks.Webhook` ]
            A list of webhook objects for the give channel.

        Raises
        ------
        :obj:`~hikari.errors.BadRequestHTTPError`
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the channel is not found.
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you either lack the ``MANAGE_WEBHOOKS`` permission or
            can not see the given channel.
        """
        payload = await self._session.get_channel_webhooks(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel))
        )
        return [webhooks.Webhook.deserialize(webhook) for webhook in payload]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities that are used to describe both custom and Unicode emojis on Discord."""
import typing

import attr

from hikari import bases
from hikari import users
from hikari.internal import marshaller

__all__ = ["Emoji", "UnicodeEmoji", "UnknownEmoji", "GuildEmoji"]


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class Emoji(bases.HikariEntity, marshaller.Deserializable):
    """Base class for all emojis."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class UnicodeEmoji(Emoji):
    """Represents a unicode emoji."""

    #: The codepoints that form the emoji.
    #:
    #: :type: :obj:`~str`
    name: str = marshaller.attrib(deserializer=str)

    @property
    def url_name(self) -> str:
        """Get the format of this emoji used in request routes."""
        return self.name

    @property
    def mention(self) -> str:
        """Get the format of this emoji used for sending it in a channel."""
        return self.name


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class UnknownEmoji(Emoji, bases.UniqueEntity):
    """Represents a unknown emoji."""

    #: The name of the emoji.
    #:
    #: :type: :obj:`~str`, optional
    name: typing.Optional[str] = marshaller.attrib(deserializer=str, if_none=None)

    #: Whether the emoji is animated.
    #:
    #: Will be :obj:`None` when received in Message Reaction Remove and Message
    #: Reaction Remove Emoji events.
    #:
    #: :type: :obj:`~bool`, optional
    is_animated: typing.Optional[bool] = marshaller.attrib(
        raw_name="animated", deserializer=bool, if_undefined=False, if_none=None, default=False
    )

    @property
    def url_name(self) -> str:
        """Get the format of this emoji used in request routes."""
        return f"{self.name}:{self.id}"


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildEmoji(UnknownEmoji):
    """Represents a guild emoji."""

    #: The whitelisted role IDs to use this emoji.
    #:
    #: :type: :obj:`~typing.Set` [ :obj:`~hikari.entities.Snowflake` ]
    role_ids: typing.Set[bases.Snowflake] = marshaller.attrib(
        raw_name="roles",
        deserializer=lambda roles: {bases.Snowflake.deserialize(r) for r in roles},
        if_undefined=set,
        factory=set,
    )

    #: The user that created the emoji.
    #:
    #: Note
    #: ----
    #: This will be :obj:`~None` if you are missing the ``MANAGE_EMOJIS``
    #: permission in the server the emoji is from.
    #:
    #:
    #: :type: :obj:`~hikari.users.User`, optional
    user: typing.Optional[users.User] = marshaller.attrib(
        deserializer=users.User.deserialize, if_none=None, if_undefined=None, default=None
    )

    #: Whether this emoji must be wrapped in colons.
    #:
    #: :type: :obj:`~bool`
    is_colons_required: bool = marshaller.attrib(raw_name="require_colons", deserializer=bool)

    #: Whether the emoji is managed by an integration.
    #:
    #: :type: :obj:`~bool`
    is_managed: bool = marshaller.attrib(raw_name="managed", deserializer=bool)

    #: Whether this emoji can currently be used, may be :obj:`False` due to
    #: a loss of Sever Boosts on the emoji's guild.
    #:
    #: :type: :obj:`~bool`
    is_available: bool = marshaller.attrib(raw_name="available", deserializer=bool)

    @property
    def mention(self) -> str:
        """Get the format of this emoji used for sending it in a channel."""
        return f"<{'a' if self.is_animated else ''}:{self.url_name}>"


def deserialize_reaction_emoji(payload: typing.Dict) -> typing.Union[UnicodeEmoji, UnknownEmoji]:
    """Deserialize a reaction emoji into an emoji."""
    if payload.get("id"):
        return UnknownEmoji.deserialize(payload)

    return UnicodeEmoji.deserialize(payload)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Logic for handling all requests to reaction endpoints."""

__all__ = ["RESTReactionComponent"]

import abc
import datetime
import typing

from hikari import bases
from hikari import channels as _channels
from hikari import emojis
from hikari import messages as _messages
from hikari import users
from hikari.clients.rest import base
from hikari.internal import pagination


class RESTReactionComponent(base.BaseRESTComponent, abc.ABC):  # pylint: disable=W0223
    """The REST client component for handling requests to reaction endpoints."""

    async def create_reaction(
        self,
        channel: bases.Hashable[_channels.Channel],
        message: bases.Hashable[_messages.Message],
        emoji: typing.Union[emojis.Emoji, str],
    ) -> None:
        """Add a reaction to the given message in the given channel.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the channel to add this reaction in.
        message : :obj:`~typing.Union` [ :obj:`~hikari.messages.Message`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the message to add the reaction in.
        emoji : :obj:`~typing.Union` [ :obj:`~hikari.emojis.Emoji`, :obj:`~str` ]
            The emoji to add. This can either be an emoji object or a string
            representation of an emoji. The string representation will be either
            ``"name:id"`` for custom emojis else it's unicode character(s)  (can
            be UTF-32).

        Raises
        ------
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If this is the first reaction using this specific emoji on this
            message and you lack the ``ADD_REACTIONS`` permission. If you lack
            ``READ_MESSAGE_HISTORY``, this may also raise this error.
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the channel or message is not found, or if the emoji is not found.
        :obj:`~hikari.errors.BadRequestHTTPError`
            If the emoji is not valid, unknown, or formatted incorrectly.
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        """
        await self._session.create_reaction(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
            message_id=str(message.id if isinstance(message, bases.UniqueEntity) else int(message)),
            emoji=str(getattr(emoji, "url_name", emoji)),
        )

    async def delete_reaction(
        self,
        channel: bases.Hashable[_channels.Channel],
        message: bases.Hashable[_messages.Message],
        emoji: typing.Union[emojis.Emoji, str],
    ) -> None:
        """Remove your own reaction from the given message in the given channel.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the channel to add this reaction in.
        message : :obj:`~typing.Union` [ :obj:`~hikari.messages.Message`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the message to add the reaction in.
        emoji : :obj:`~typing.Union` [ :obj:`~hikari.emojis.Emoji`, :obj:`~str` ]
            The emoji to add. This can either be an emoji object or a
            string representation of an emoji. The string representation will be
            either ``"name:id"`` for custom emojis else it's unicode
            character(s) (can be UTF-32).

        Raises
        ------
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If this is the first reaction using this specific emoji on this
            message and you lack the ``ADD_REACTIONS`` permission. If you lack
            ``READ_MESSAGE_HISTORY``, this may also raise this error.
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the channel or message is not found, or if the emoji is not
            found.
        :obj:`~hikari.errors.BadRequestHTTPError`
            If the emoji is not valid, unknown, or formatted incorrectly.
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        """
        await self._session.delete_own_reaction(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
            message_id=str(message.id if isinstance(message, bases.UniqueEntity) else int(message)),
            emoji=str(getattr(emoji, "url_name", emoji)),
        )

    async def delete_all_reactions(
        self, channel: bases.Hashable[_channels.Channel], message: bases.Hashable[_messages.Message],
    ) -> None:
        """Delete all reactions from a given message in a given channel.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the channel to get the message from.
        message : :obj:`~typing.Union` [ :obj:`~hikari.messages.Message`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the message to remove all reactions from.

        Raises
        ------
        :obj:`~hikari.errors.BadRequestHTTPError`
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the channel or message is not found.
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you lack the ``MANAGE_MESSAGES`` permission.
        """
        await self._session.delete_all_reactions(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
            message_id=str(message.id if isinstance(message, bases.UniqueEntity) else int(message)),
        )

    async def delete_all_reactions_for_emoji(
        self,
        channel: bases.Hashable[_channels.Channel],
        message: bases.Hashable[_messages.Message],
        emoji: typing.Union[emojis.Emoji, str],
    ) -> None:
        """Remove all reactions for a single given emoji on a given message.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the channel to get the message from.
        message : :obj:`~typing.Union` [ :obj:`~hikari.messages.Message`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the message to delete the reactions from.
        emoji : :obj:`~typing.Union` [ :obj:`~hikari.emojis.Emoji`, :obj:`~str` ]
            The object or string representatiom of the emoji to delete. The
            string representation will be either ``"name:id"`` for custom emojis
            else it's unicode character(s) (can be UTF-32).

        Raises
        ------
        :obj:`~hikari.errors.BadRequestHTTPError`
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the channel or message or emoji or user is not found.
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you lack the ``MANAGE_MESSAGES`` permission, or the channel is a
            DM channel.
        """
        await self._session.delete_all_reactions_for_emoji(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
            message_id=str(message.id if isinstance(message, bases.UniqueEntity) else int(message)),
            emoji=str(getattr(emoji, "url_name", emoji)),
        )

    def fetch_reactors_after(
        self,
        channel: bases.Hashable[_channels.Channel],
        message: bases.Hashable[_messages.Message],
        emoji: typing.Union[emojis.Emoji, str],
        *,
        after: typing.Union[datetime.datetime, bases.Hashable[users.User]] = 0,
        limit: typing.Optional[int] = None,
    ) -> typing.AsyncIterator[users.User]:
        """Get an async iterator of the users who reacted to a message.

        This returns the users created after a given user object/ID or from the
        oldest user who reacted.

        Parameters
        ----------
        channel : :obj:`~typing.Union` [ :obj:`~hikari.channels.Channel`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the channel to get the message from.
        message : :obj:`~typing.Union` [ :obj:`~hikari.messages.Message`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the message to get the reactions from.
        emoji : :obj:`~typing.Union` [ :obj:`~hikari.emojis.Emoji`, :obj:`~str` ]
            The emoji to get. This can either be it's object or the string
            representation of the emoji. The string representation will be
            either ``"name:id"`` for custom emojis else it's unicode
            character(s) (can be UTF-32).
        after : :obj:`~typing.Union` [ :obj:`~datetime.datetime`, :obj:`~hikari.users.User`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            If specified, a object or ID user. If specified, only users with a
            snowflake that is lexicographically greater than the value will be
            returned.
        limit : :obj:`~str`
            If specified, the limit of the number of users this iterator should
            return.

        Example
        -------
        .. code-block:: python

            async for user in client.fetch_reactors_after(channel, message, emoji, after=9876543, limit=1231):
                if user.is_bot:
                    await client.kick_member(channel.guild_id, user)

        Returns
        -------
        :obj:`~typing.AsyncIterator` [ :obj:`~hikari.users.User` ]
            An async iterator of user objects.

        Raises
        ------
        :obj:`~hikari.errors.BadRequestHTTPError`
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you lack access to the message.
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the channel or message is not found.
        """
        if isinstance(after, datetime.datetime):
            after = str(bases.Snowflake.from_datetime(after))
        else:
            after = str(after.id if isinstance(after, bases.UniqueEntity) else int(after))
        return pagination.pagination_handler(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
            message_id=str(message.id if isinstance(message, bases.UniqueEntity) else int(message)),
            emoji=getattr(emoji, "url_name", emoji),
            deserializer=users.User.deserialize,
            direction="after",
            request=self._session.get_reactions,
            reversing=False,
            start=after,
            limit=limit,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""The logic for handling requests to gateway endpoints."""

__all__ = ["RESTGatewayComponent"]

import abc

from hikari import gateway_entities
from hikari.clients.rest import base


class RESTGatewayComponent(base.BaseRESTComponent, abc.ABC):  # pylint: disable=W0223
    """The REST client component for handling requests to gateway endpoints."""

    async def fetch_gateway_url(self) -> str:
        """Get a generic url used for establishing a Discord gateway connection.

        Returns
        -------
        :obj:`~str`
            A static URL to use to connect to the gateway with.

        Note
        ----
        Users are expected to attempt to cache this result.
        """
        return await self._session.get_gateway()

    async def fetch_gateway_bot(self) -> gateway_entities.GatewayBot:
        """Get bot specific gateway information.

        Returns
        -------
        :obj:`~hikari.gateway_entities.GatewayBot`
            The bot specific gateway information object.

        Note
        ----
        Unlike :meth:`fetch_gateway_url`, this requires a valid token to work.
        """
        payload = await self._session.get_gateway_bot()
        return gateway_entities.GatewayBot.deserialize(payload)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""The logic for handling all requests to user endpoints."""

__all__ = ["RESTUserComponent"]

import abc

from hikari import bases
from hikari import users
from hikari.clients.rest import base


class RESTUserComponent(base.BaseRESTComponent, abc.ABC):  # pylint: disable=W0223
    """The REST client component for handling requests to user endpoints."""

    async def fetch_user(self, user: bases.Hashable[users.User]) -> users.User:
        """Get a given user.

        Parameters
        ----------
        user : :obj:`~typing.Union` [ :obj:`~hikari.users.User`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ]
            The object or ID of the user to get.

        Returns
        -------
        :obj:`~hikari.users.User`
            The requested user object.

        Raises
        ------
        :obj:`~hikari.errors.BadRequestHTTPError`
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the user is not found.
        """
        payload = await self._session.get_user(
            user_id=str(user.id if isinstance(user, bases.UniqueEntity) else int(user))
        )
        return users.User.deserialize(payload)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""The logic for handling requests to invite endpoints."""

__all__ = ["RESTInviteComponent"]

import abc
import typing

from hikari import invites
from hikari.clients.rest import base


class RESTInviteComponent(base.BaseRESTComponent, abc.ABC):  # pylint: disable=W0223
    """The REST client component for handling requests to invite endpoints."""

    async def fetch_invite(
        self, invite: typing.Union[invites.Invite, str], *, with_counts: bool = ...
    ) -> invites.Invite:
        """Get the given invite.

        Parameters
        ----------
        invite : :obj:`~typing.Union` [ :obj:`~hikari.invites.Invite`, :obj:`~str` ]
            The object or code of the wanted invite.
        with_counts : :bool:
            If specified, whether to attempt to count the number of
            times the invite has been used.

        Returns
        -------
        :obj:`~hikari.invites.Invite`
            The requested invite object.

        Raises
        ------
        :obj:`~hikari.errors.BadRequestHTTPError`
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the invite is not found.
        """
        payload = await self._session.get_invite(invite_code=getattr(invite, "code", invite), with_counts=with_counts)
        return invites.Invite.deserialize(payload)

    async def delete_invite(self, invite: typing.Union[invites.Invite, str]) -> None:
        """Delete a given invite.

        Parameters
        ----------
        invite : :obj:`~typing.Union` [ :obj:`~hikari.invites.Invite`, :obj:`~str` ]
            The object or ID for the invite to be deleted.

        Returns
        -------
        :obj:`~None`
            Nothing, unlike what the API specifies. This is done to maintain
            consistency with other calls of a similar nature in this API wrapper.

        Raises
        ------
        :obj:`~hikari.errors.NotFoundHTTPError`
            If the invite is not found.
        :obj:`~hikari.errors.ForbiddenHTTPError`
            If you lack either ``MANAGE_CHANNELS`` on the channel the invite
            belongs to or ``MANAGE_GUILD`` for guild-global delete.
        """
        await self._session.delete_invite(invite_code=getattr(invite, "code", invite))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Asyncio extensions and utilities.

|internal|
"""
from __future__ import annotations

__all__ = ["completed_future"]

import asyncio
import typing

from hikari.internal import more_typing


@typing.overload
def completed_future() -> more_typing.Future[None]:
    """Return a completed future with no result."""


@typing.overload
def completed_future(result: more_typing.T_contra, /) -> more_typing.Future[more_typing.T_contra]:
    """Return a completed future with the given value as the result."""


def completed_future(result=None, /):
    """Create a future on the current running loop that is completed, then return it.

    Parameters
    ----------
    result : :obj:`~typing.Any`
        The value to set for the result of the future.

    Returns
    -------
    :obj:`~asyncio.Future`
        The completed future.
    """
    future = asyncio.get_event_loop().create_future()
    future.set_result(result)
    return future


def wait(
    aws: typing.Union[more_typing.Coroutine, typing.Awaitable], *, timeout=None, return_when=asyncio.ALL_COMPLETED
) -> more_typing.Coroutine[typing.Tuple[typing.Set[more_typing.Future], typing.Set[more_typing.Future]]]:
    """Run awaitable objects in the aws set concurrently.

    This blocks until the condition specified by `return_value`.

    Returns
    -------
    :obj:`~typing.Tuple` with two :obj:`~typing.Set` of futures.
        The coroutine returned by :obj:`~asyncio.wait` of two sets of
        Tasks/Futures (done, pending).
    """
    # noinspection PyTypeChecker
    return asyncio.wait([asyncio.ensure_future(f) for f in aws], timeout=timeout, return_when=return_when)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Basic transformation utilities."""
__all__ = [
    "nullable_cast",
    "try_cast",
    "try_cast_or_defer_unary_operator",
    "put_if_specified",
    "image_bytes_to_image_data",
    "parse_http_date",
    "parse_iso_8601_ts",
    "discord_epoch_to_datetime",
    "unix_epoch_to_datetime",
    "Seekable",
    "make_resource_seekable",
    "pluralize",
    "snoop_typehint_from_scope",
    "FileLikeT",
    "BytesLikeT",
]

import base64
import contextlib
import datetime
import email.utils
import enum
import functools
import io
import operator
import re
import types
import typing

IntFlagT = typing.TypeVar("IntFlagT", bound=enum.IntFlag)
RawIntFlagValueT = typing.Union[typing.AnyStr, typing.SupportsInt, int]
DISCORD_EPOCH: typing.Final[int] = 1_420_070_400
ISO_8601_DATE_PART: typing.Final[typing.Pattern] = re.compile(r"^(\d{4})-(\d{2})-(\d{2})")
ISO_8601_TIME_PART: typing.Final[typing.Pattern] = re.compile(r"T(\d{2}):(\d{2}):(\d{2})(?:\.(\d{1,6}))?", re.I)
ISO_8601_TZ_PART: typing.Final[typing.Pattern] = re.compile(r"([+-])(\d{2}):(\d{2})$")

CastInputT = typing.TypeVar("CastInputT")
CastOutputT = typing.TypeVar("CastOutputT")
DefaultT = typing.TypeVar("DefaultT")
TypeCastT = typing.Callable[[CastInputT], CastOutputT]
ResultT = typing.Union[CastOutputT, DefaultT]
BytesLikeT = typing.Union[bytes, bytearray, memoryview, str, io.StringIO, io.BytesIO]
FileLikeT = typing.Union[BytesLikeT, io.BufferedRandom, io.BufferedReader, io.BufferedRWPair]


def nullable_cast(value: CastInputT, cast: TypeCastT, /) -> ResultT:
    """Attempt to cast the given ``value`` with the given ``cast``.

    This will only succeed if ``value`` is not :obj:`~None`. If it is
    :obj:`~None`, then :obj:`~None` is returned instead.
    """
    if value is None:
        return None
    return cast(value)


def try_cast(value: CastInputT, cast: TypeCastT, default: DefaultT = None, /) -> ResultT:
    """Try to cast the given value to the given cast.

    If it throws a :obj:`~Exception` or derivative, it will return ``default``
    instead of the cast value instead.
    """
    with contextlib.suppress(Exception):
        return cast(value)
    return default


def try_cast_or_defer_unary_operator(type_: typing.Type, /):
    """Return a unary operator that will try to cast the given input to the type provided.

    Parameters
    ----------
    type_ : :obj:`~typing.Callable` [ ..., ``output type`` ]
        The type to cast to.
    """
    return lambda data: try_cast(data, type_, data)


def put_if_specified(
    mapping: typing.Dict[typing.Hashable, typing.Any],
    key: typing.Hashable,
    value: typing.Any,
    type_after: typing.Optional[TypeCastT] = None,
    /,
) -> None:
    """Add a value to the mapping under the given key as long as the value is not ``...``.

    Parameters
    ----------
    mapping : :obj:`~typing.Dict` [ :obj:`~typing.Hashable`, :obj:`~typing.Any` ]
        The mapping to add to.
    key : :obj:`~typing.Hashable`
        The key to add the value under.
    value : :obj:`~typing.Any`
        The value to add.
    type_after : :obj:`~typing.Callable` [ [ ``input type`` ], ``output type`` ], optional
        Type to apply to the value when added.
    """
    if value is not ...:
        if type_after:
            mapping[key] = type_after(value)
        else:
            mapping[key] = value


def image_bytes_to_image_data(img_bytes: typing.Optional[bytes] = None, /) -> typing.Optional[str]:
    """Encode image bytes into an image data string.

    Parameters
    ----------
    img_bytes : :obj:`~bytes`, optional
        The image bytes.

    Raises
    ------
    :obj:`~ValueError`
        If the image type passed is not supported.

    Returns
    -------
    :obj:`~str`, optional
        The ``image_bytes`` given encoded into an image data string or
        :obj:`~None`.

    Note
    ----
    Supported image types: ``.png``, ``.jpeg``, ``.jfif``, ``.gif``, ``.webp``
    """
    if img_bytes is None:
        return None

    if img_bytes[:8] == b"\211PNG\r\n\032\n":
        img_type = "image/png"
    elif img_bytes[6:10] in (b"Exif", b"JFIF"):
        img_type = "image/jpeg"
    elif img_bytes[:6] in (b"GIF87a", b"GIF89a"):
        img_type = "image/gif"
    elif img_bytes.startswith(b"RIFF") and img_bytes[8:12] == b"WEBP":
        img_type = "image/webp"
    else:
        raise ValueError("Unsupported image type passed")

    image_data = base64.b64encode(img_bytes).decode()

    return f"data:{img_type};base64,{image_data}"


def parse_http_date(date_str: str, /) -> datetime.datetime:
    """Return the HTTP date as a datetime object.

    Parameters
    ----------
    date_str : :obj:`~str`
        The RFC-2822 (section 3.3) compliant date string to parse.

    Returns
    -------
    :obj:`~datetime.datetime`
        The HTTP date as a datetime object.

    See Also
    --------
    `<https://www.ietf.org/rfc/rfc2822.txt>`_
    """
    return email.utils.parsedate_to_datetime(date_str)


def parse_iso_8601_ts(date_string: str, /) -> datetime.datetime:
    """Parse an ISO 8601 date string into a :obj:`~datetime.datetime` object.

    Parameters
    ----------
    date_string : :obj:`~str`
        The ISO 8601 compliant date string to parse.

    Returns
    -------
    :obj:`~datetime.datetime`
        The ISO 8601 date string as a datetime object.

    See Also
    --------
    `<https://en.wikipedia.org/wiki/ISO_8601>`_
    """
    year, month, day = map(int, ISO_8601_DATE_PART.findall(date_string)[0])

    time_part = ISO_8601_TIME_PART.findall(date_string)[0]
    hour, minute, second, partial = time_part

    # Pad the millisecond part if it is not in microseconds, otherwise Python will complain.
    partial = partial + (6 - len(partial)) * "0"
    hour, minute, second, partial = int(hour), int(minute), int(second), int(partial)
    if date_string.endswith(("Z", "z")):
        timezone = datetime.timezone.utc
    else:
        sign, tz_hour, tz_minute = ISO_8601_TZ_PART.findall(date_string)[0]
        tz_hour, tz_minute = int(tz_hour), int(tz_minute)
        offset = datetime.timedelta(hours=tz_hour, minutes=tz_minute)
        if sign == "-":
            offset = -offset
        timezone = datetime.timezone(offset)

    return datetime.datetime(year, month, day, hour, minute, second, partial, timezone)


def discord_epoch_to_datetime(epoch: int, /) -> datetime.datetime:
    """Parse a Discord epoch into a :obj:`~datetime.datetime` object.

    Parameters
    ----------
    epoch : :obj:`~int`
        Number of milliseconds since 1/1/2015 (UTC)

    Returns
    -------
    :obj:`~datetime.datetime`
        Number of seconds since 1/1/1970 within a datetime object (UTC).
    """
    return datetime.datetime.fromtimestamp(epoch / 1000 + DISCORD_EPOCH, datetime.timezone.utc)


def unix_epoch_to_datetime(epoch: int, /) -> datetime.datetime:
    """Parse a UNIX epoch to a :obj:`~datetime.datetime` object.

    Parameters
    ----------
    epoch : :obj:`~int`
        Number of milliseconds since 1/1/1970 (UTC)

    Returns
    -------
    :obj:`~datetime.datetime`
        Number of seconds since 1/1/1970 within a datetime object (UTC).
    """
    return datetime.datetime.fromtimestamp(epoch / 1000, datetime.timezone.utc)


class Seekable(typing.Protocol[typing.AnyStr]):
    """Structural type for an IO object that supports seek operations."""

    def seek(
        self, offset: int, whence: typing.Union[typing.Literal[0], typing.Literal[1], typing.Literal[2]] = 0, /
    ) -> None:
        """Seek to the given offset.

        Parameters
        ----------
        offset : :obj:`~int`
            The offset to seek to.
        whence : :obj:`~int`
            If ``0``, as the default, then use absolute file positioning.
            If ``1``, then seek to the current position.
            If ``2``, then seek relative to the end of the file.
        """

    def tell(self) -> int:
        """Return the stream position.

        Returns
        -------
        :obj:`~int`
            The stream position.
        """

    def read(self) -> typing.AnyStr:
        """Read part of a string.

        Returns
        -------
        :obj:`~str`
            The string that was read.
        """

    def close(self) -> None:
        """Close the stream."""


def make_resource_seekable(resource: typing.Any, /) -> Seekable:
    """Make a seekable resource to use off some representation of data.

    This supports :obj:`~bytes`, :obj:`~bytearray`, :obj:`~memoryview`, and
    :obj:`~str`. Anything else is just returned.

    Parameters
    ----------
    resource : :obj:`~typing.Any`
        The resource to check.

    Returns
    -------
    :obj:`~typing.Union` [ :obj:`~io.BytesIO`, :obj:`~io.StringIO` ]
        An stream-compatible resource where possible.
    """
    if isinstance(resource, (bytes, bytearray)):
        resource = io.BytesIO(resource)
    elif isinstance(resource, memoryview):
        resource = io.BytesIO(resource.tobytes())
    elif isinstance(resource, str):
        resource = io.StringIO(resource)

    return resource


def get_bytes_from_resource(resource: typing.Any) -> bytes:
    """Take in any file-like object and return the raw bytes data from it.

    Supports any ``FileLikeT`` type that isn't string based.
    Anything else is just returned.

    Parameters
    ----------
    resource : ``FileLikeT``
        The resource to get bytes from.

    Returns
    -------
    :obj:`~bytes`
        The resulting bytes.
    """
    if isinstance(resource, bytearray):
        resource = bytes(resource)
    elif isinstance(resource, memoryview):
        resource = resource.tobytes()
    # Targets the io types found in FileLikeT and BytesLikeT
    elif hasattr(resource, "read"):
        resource = resource.read()

    return resource


def pluralize(count: int, name: str, suffix: str = "s") -> str:
    """Pluralizes a word."""
    return f"{count} {name + suffix}" if count - 1 else f"{count} {name}"


def snoop_typehint_from_scope(frame: types.FrameType, typehint: typing.Union[str, typing.Any]) -> typing.Any:
    """Resolve a string type hint from a given stack frame.

    This snoops around the local and global scope for the given frame to find
    the given attribute name, taking into account nested function calls. The
    reason to do this is that if a string type hint is used, or the
    ``from __future__ import annotations`` directive is used, the physical thing
    that the type hint represents will no longer be evaluated by the
    interpreter. This is an implementation that does not require the use of
    :obj:`~eval`, and thus reduces the risk of arbitrary code execution as a
    result.

    Nested parameters such as :obj:`~typing.Sequence` should also be able to be
    resolved correctly.

    Parameters
    ----------
    frame : :obj:`~types.FrameType`
        The stack frame that the element with the typehint was defined in.
        This is retrieved using :obj:`~inspect.stack` ``(frame_no)[0][0]``,
        where ``frame_no`` is the number of frames from this invocation that
        you want to snoop the scope at.
    typehint : :obj:`~typing.Union` [ :obj:`~str`, :obj:`~typing.Any` ]
        The type hint to resolve. If a non-:obj:`~str` is passed, then this is
        returned immediately as the result.

    Returns
    -------
    :obj:`~typing.Any`
        The physical representation of the given type hint.

    Raises
    ------
    :obj:`~NameError`
        If the attribute was not found.

    Warnings
    --------
    The input frame must be manually dereferenced using the ``del`` keyword
    after use. Any functions that are decorated and wrapped when using this
    lookup must use :obj:`~functools.wraps` to ensure that the correct scope is
    identified on the stack.

    This is incredibly unpythonic and baremetal, but due to
    `PEP 563 <https://www.python.org/dev/peps/pep-0563/>` there is no other
    consistent way of making this work correctly.
    """
    if not isinstance(typehint, str):
        return typehint

    fragments = typehint.split(".")

    for scope in (frame.f_locals, frame.f_globals):
        try:
            scope = scope[fragments[0]]
            for attr in fragments[1:]:
                scope = getattr(scope, attr)
            return scope
        except (AttributeError, KeyError):
            pass
    raise NameError(f"No attribute {typehint} was found in enclosing scope")


def dereference_int_flag(
    int_flag_type: typing.Type[IntFlagT],
    raw_value: typing.Union[RawIntFlagValueT, typing.Collection[RawIntFlagValueT]],
) -> IntFlagT:
    """Cast to the provided :obj:`~enum.IntFlag` type.

    This supports resolving bitfield integers as well as decoding a sequence
    of case insensitive flag names into one combined value.

    Parameters
    ----------
    int_flag_type : :obj:`~typing.Type` [ :obj:`~enum.IntFlag` ]
        The type of the int flag to check.
    raw_value : ``Castable Value``
        The raw value to convert.

    Returns
    -------
    :obj:`~enum.IntFlag`
        The cast value as a flag.

    Notes
    -----
    Types that are a ``Castable Value`` include:
    - :obj:`~str`
    - :obj:`~int`
    - :obj:`~typing.SupportsInt`
    - :obj:`~typing.Collection` [ ``Castable Value`` ]

    When a collection is passed, values will be combined using functional
    reduction via the :obj:operator.or_` operator.
    """
    if isinstance(raw_value, str) and raw_value.isdigit():
        raw_value = int(raw_value)

    if not isinstance(raw_value, int):
        raw_value = functools.reduce(operator.or_, (int_flag_type[name.upper()] for name in raw_value))

    return int_flag_type(raw_value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""The logic for handling all requests to voice endpoints."""

__all__ = ["RESTVoiceComponent"]

import abc
import typing

from hikari import voices
from hikari.clients.rest import base


class RESTVoiceComponent(base.BaseRESTComponent, abc.ABC):  # pylint: disable=W0223
    """The REST client component for handling requests to voice endpoints."""

    async def fetch_voice_regions(self) -> typing.Sequence[voices.VoiceRegion]:
        """Get the voice regions that are available.

        Returns
        -------
        :obj:`~typing.Sequence` [ :obj:`~hikari.voices.VoiceRegion` ]
            A list of voice regions available

        Note
        ----
        This does not include VIP servers.
        """
        payload = await self._session.list_voice_regions()
        return [voices.VoiceRegion.deserialize(region) for region in payload]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Network components for the Hikari Discord API.

These components describe the low level parts of Hikari. No model classes exist
for these; the majority of communication is done via JSON arrays and objects.
"""
from hikari.net import codes
from hikari.net import ratelimits
from hikari.net import rest
from hikari.net import routes
from hikari.net import shards
from hikari.net import user_agents

from hikari.net.codes import *
from hikari.net.rest import *
from hikari.net.shards import *

__all__ = codes.__all__ + shards.__all__ + rest.__all__



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Various functional types and metatypes.

|internal|
"""
__all__ = ["SingletonMeta", "Singleton"]

import abc
import inspect
import typing

from hikari.internal import more_collections


class SingletonMeta(type):
    """Metaclass that makes the class a singleton.

    Once an instance has been defined at runtime, it will exist until the
    interpreter that created it is terminated.

    Example
    -------
    .. code-block:: python

        >>> class Unknown(metaclass=SingletonMeta):
        ...     def __init__(self):
        ...         print("Initialized an Unknown!")

        >>> Unknown() is Unknown()    # True

    Note
    ----
    The constructors of instances of this metaclass must not take any arguments
    other than ``self``.

    Warning
    -------
    Constructing instances of class instances of this metaclass may not be
    thread safe.
    """

    # pylint: disable=E1136
    ___instance_dict_t___ = more_collections.WeakKeyDictionary[typing.Type[typing.Any], typing.Any]
    # pylint: enable=E1136
    ___instances___: ___instance_dict_t___ = more_collections.WeakKeyDictionary()
    __slots__ = ()

    def __call__(cls):
        if cls not in SingletonMeta.___instances___:
            SingletonMeta.___instances___[cls] = super().__call__()
        return SingletonMeta.___instances___[cls]


class Singleton(metaclass=SingletonMeta):
    """Base type for anything implementing the :obj:`~SingletonMeta` metaclass.

    Once an instance has been defined at runtime, it will exist until the
    interpreter that created it is terminated.

    Example
    -------
    .. code-block:: python

        >>> class MySingleton(Singleton):
        ...    pass

        >>> assert MySingleton() is MySingleton()

    Note
    ----
    The constructors of child classes must not take any arguments other than
    ``self``.

    Warning
    -------
    Constructing instances of this class or derived classes may not be thread
    safe.
    """


class UniqueFunctionMeta(abc.ABCMeta):
    """Metaclass for mixins that are expected to provide unique function names.

    If subclassing from two mixins that are derived from this type and both
    mixins provide the same function, a type error is raised when the class is
    defined.

    Note
    ----
    This metaclass derives from :obj:`~abc.ABCMeta`, and thus is compatible
    with abstract method conduit.
    """

    @classmethod
    def __prepare__(mcs, name, bases, **kwargs):
        routines = {}

        for base in bases:
            for identifier, method in inspect.getmembers(base, inspect.isroutine):
                if identifier.startswith("__"):
                    continue

                if identifier in routines and method != routines[identifier]:
                    raise TypeError(
                        f"Conflicting methods {routines[identifier].__qualname__} and {method.__qualname__} found."
                    )

                routines[identifier] = method

        return super().__prepare__(name, bases, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Provides a base for any type of websocket client."""

__all__ = ["RunnableClient"]

import abc
import asyncio
import contextlib
import logging
import signal
import typing


class RunnableClient(abc.ABC):
    """Base for any websocket client that must be kept alive."""

    __slots__ = ("logger",)

    #: The logger to use for this client.
    #:
    #: :type: :obj:`~logging.Logger`
    logger: logging.Logger

    @abc.abstractmethod
    def __init__(self, logger: typing.Union[logging.Logger, logging.LoggerAdapter]) -> None:
        self.logger = logger

    @abc.abstractmethod
    async def start(self) -> None:  # noqa: D401
        """Start the component."""

    @abc.abstractmethod
    async def close(self) -> None:
        """Shut down the component."""

    @abc.abstractmethod
    async def join(self) -> None:
        """Wait for the component to terminate."""

    def run(self) -> None:
        """Execute this component on an event loop.

        Performs the same job as :meth:`start`, but provides additional
        preparation such as registering OS signal handlers for interrupts,
        and preparing the initial event loop.

        This enables the client to be run immediately without having to
        set up the :mod:`asyncio` event loop manually first.
        """
        loop = asyncio.get_event_loop()

        def sigterm_handler(*_):
            raise KeyboardInterrupt()

        ex = None

        try:
            with contextlib.suppress(NotImplementedError):
                # Not implemented on Windows
                loop.add_signal_handler(signal.SIGTERM, sigterm_handler)

            loop.run_until_complete(self.start())
            loop.run_until_complete(self.join())

            self.logger.info("client has shut down")

        except KeyboardInterrupt as _ex:
            self.logger.info("received signal to shut down client")
            loop.run_until_complete(self.close())
            # Apparently you have to alias except clauses or you get an
            # UnboundLocalError.
            ex = _ex
        finally:
            loop.run_until_complete(self.close())
            with contextlib.suppress(NotImplementedError):
                # Not implemented on Windows
                loop.remove_signal_handler(signal.SIGTERM)

        if ex:
            raise ex from ex



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Represents various forms of media such as images."""
__all__ = ["TextIO", "BytesIO", "IO", "safe_read_file"]

import io
import typing

import aiofiles

from hikari.internal import conversions

TextIO = typing.Union[aiofiles.threadpool.text.AsyncTextIOWrapper, typing.TextIO, io.StringIO, str]

BytesIO = typing.Union[
    aiofiles.threadpool.binary.AsyncBufferedIOBase,
    aiofiles.threadpool.binary.AsyncBufferedReader,
    aiofiles.threadpool.binary.AsyncFileIO,
    typing.BinaryIO,
    io.BytesIO,
    bytes,
    bytearray,
    memoryview,
]

IO = typing.Union[TextIO, BytesIO]


async def safe_read_file(file: IO) -> typing.Tuple[str, conversions.FileLikeT]:
    """Safely read an ``IO`` like object."""
    raise NotImplementedError  # TODO: Nekokatt: update this.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Logic for handling Discord's generic paginated endpoints.

|internal|
"""

__all__ = ["pagination_handler"]

import typing

from hikari.internal import more_typing


async def pagination_handler(
    deserializer: typing.Callable[[typing.Any], typing.Any],
    direction: typing.Union[typing.Literal["before"], typing.Literal["after"]],
    request: typing.Callable[..., more_typing.Coroutine[typing.Any]],
    reversing: bool,
    start: typing.Union[str, None],
    limit: typing.Optional[int] = None,
    id_getter: typing.Callable[[typing.Any], str] = lambda entity: str(entity.id),
    **kwargs,
) -> typing.AsyncIterator[typing.Any]:
    """Generate an async iterator for handling paginated endpoints.

    This will handle Discord's ``before`` and ``after`` pagination.

    Parameters
    ----------
    deserializer : :obj:`~typing.Callable` [ [ :obj:`~typing.Any` ], :obj:`~typing.Any` ]
        The deserializer to use to deserialize raw elements.
    direction : :obj:`~typing.Union` [ ``"before"``, ``"after"`` ]
        The direction that this paginator should go in.
    request : :obj:`~typing.Callable` [ ``...``, :obj:`~typing.Coroutine` [ :obj:`~typing.Any`, :obj:`~typing.Any`, :obj:`~typing.Any` ] ]
        The :obj:`hikari.net.rest_sessions.LowLevelRestfulClient` method that should be
        called to make requests for this paginator.
    reversing : :obj:`~bool`
        Whether the retrieved array of objects should be reversed before
        iterating through it, this is needed for certain endpoints like
        ``fetch_messages_before`` where the order is static regardless of
        if you're using ``before`` or ``after``.
    start : :obj:`~int`, optional
        The snowflake ID that this paginator should start at, ``0`` may be
        passed for ``forward`` pagination to start at the first created
        entity and :obj:`~None` may be passed for ``before`` pagination to
        start at the newest entity (based on when it's snowflake timestamp).
    limit : :obj:`~int`, optional
        The amount of deserialized entities that the iterator should return
        total, will be unlimited if set to :obj:`~None`.
    id_getter : :obj:`~typing.Callable` [ [ :obj:`~typing.Any` ], :obj:`~str` ]
    **kwargs
        Kwargs to pass through to ``request`` for every request made along
        with the current decided limit and direction snowflake.

    Returns
    -------
    :obj:`~typing.AsyncIterator` [ :obj:`~typing.Any` ]
        An async iterator of the found deserialized found objects.

    """
    while payloads := await request(
        limit=100 if limit is None or limit > 100 else limit,
        **{direction: start if start is not None else ...},
        **kwargs,
    ):
        if reversing:
            payloads.reverse()
        if limit is not None:
            limit -= len(payloads)

        for payload in payloads:
            entity = deserializer(payload)
            yield entity
        if limit == 0:
            break
        # TODO: @FasterSpeeding: can `payloads` ever be empty, leading this to be undefined?
        start = id_getter(entity)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Logic for generating a allowed mentions dict objects.

|internal|
"""

__all__ = ["generate_allowed_mentions"]

import typing

from hikari import bases
from hikari import guilds
from hikari import users
from hikari.internal import assertions
from hikari.internal import more_collections


def generate_allowed_mentions(
    mentions_everyone: bool,
    user_mentions: typing.Union[typing.Collection[bases.Hashable[users.User]], bool],
    role_mentions: typing.Union[typing.Collection[bases.Hashable[guilds.GuildRole]], bool],
) -> typing.Dict[str, typing.Sequence[str]]:
    """Generate an allowed mentions object based on input mention rules.

    Parameters
    ----------
    mentions_everyone : :obj:`~bool`
        Whether ``@everyone`` and ``@here`` mentions should be resolved by
        discord and lead to actual pings.
    user_mentions : :obj:`~typing.Union` [ :obj:`~typing.Collection` [ :obj:`~typing.Union` [ :obj:`~hikari.users.User`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ], :obj:`~bool` ]
        Either an array of user objects/IDs to allow mentions for,
        :obj:`~True` to allow all user mentions or :obj:`~False` to block all
        user mentions from resolving.
    role_mentions : :obj:`~typing.Union` [ :obj:`~typing.Collection` [ :obj:`~typing.Union` [ :obj:`~hikari.guilds.GuildRole`, :obj:`~hikari.entities.Snowflake`, :obj:`~int` ] ], :obj:`~bool` ]
        Either an array of guild role objects/IDs to allow mentions for,
        :obj:`~True` to allow all role mentions or :obj:`~False` to block all
        role mentions from resolving.

    Returns
    -------
    :obj:`~typing.Dict` [ :obj:`~str`, :obj:`~typing.Sequence` [ :obj:`~str` ] ]
        The resulting allowed mentions dict object.

    Raises
    ------
    :obj:`~ValueError`
        If more than 100 unique objects/entities are passed for
        ``role_mentions`` or ``user_mentions.
    """
    parsed_mentions = []
    allowed_mentions = {}
    if mentions_everyone is True:
        parsed_mentions.append("everyone")
    if user_mentions is True:
        parsed_mentions.append("users")
    # This covers both `False` and an array of IDs/objs by using `user_mentions or EMPTY_SEQUENCE`, where a
    # resultant empty list will mean that all user mentions are blacklisted.
    else:
        allowed_mentions["users"] = list(
            # dict.fromkeys is used to remove duplicate entries that would cause discord to return an error.
            dict.fromkeys(
                str(user.id if isinstance(user, bases.UniqueEntity) else int(user))
                for user in user_mentions or more_collections.EMPTY_SEQUENCE
            )
        )
        assertions.assert_that(len(allowed_mentions["users"]) <= 100, "Only up to 100 users can be provided.")
    if role_mentions is True:
        parsed_mentions.append("roles")
    # This covers both `False` and an array of IDs/objs by using `user_mentions or EMPTY_SEQUENCE`, where a
    # resultant empty list will mean that all role mentions are blacklisted.
    else:
        allowed_mentions["roles"] = list(
            # dict.fromkeys is used to remove duplicate entries that would cause discord to return an error.
            dict.fromkeys(
                str(role.id if isinstance(role, bases.UniqueEntity) else int(role))
                for role in role_mentions or more_collections.EMPTY_SEQUENCE
            )
        )
        assertions.assert_that(len(allowed_mentions["roles"]) <= 100, "Only up to 100 roles can be provided.")
    allowed_mentions["parse"] = parsed_mentions
    # As a note, discord will also treat an empty `allowed_mentions` object as if it wasn't passed at all, so we
    # want to use empty lists for blacklisting elements rather than just not including blacklisted elements.
    return allowed_mentions
